import matplotlib
from autobahn.twisted.util import sleep as dsleep
from twisted.internet import protocol, reactor, threads
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.logger import FileLogObserver, Logger
from twisted.python.failure import Failure

matplotlib.use('Agg')
import argparse
import os
import sys
import time
from typing import Optional

import numpy as np
import sodetlib as sdl
from ocs import ocs_agent, site_config
from ocs.ocs_agent import log_formatter
from ocs.ocs_twisted import TimeoutLock
from sodetlib.det_config import DetConfig
from sodetlib.operations import (bias_dets, bias_steps, iv, uxm_relock,
                                 uxm_setup)

NBIASLINES = 12


class PysmurfScriptProtocol(protocol.ProcessProtocol):
    """
    The process protocol used to dispatch external Pysmurf scripts, and manage
    the stdin, stdout, and stderr pipelines.

    Arguments
    ---------
    path : str
        Path of script to run.
    log : txaio.tx.Logger
        txaio logger object, used to log stdout and stderr messages.

    Attributes
    -----------
    path : str
        Path of script to run.
    log : txaio.tx.Logger
        txaio logger object, used to log stdout and stderr messages.
    end_status : twisted.python.failure.Failure
        Reason that the process ended.
    """

    def __init__(self, path, log=None):
        self.path = path
        self.log = log
        self.end_status: Optional[Failure] = None

    def connectionMade(self):
        """Called when process is started"""
        self.transport.closeStdin()

    def outReceived(self, data):
        """Called whenever data is received through stdout"""
        if self.log:
            self.log.info("{path}: {data}",
                          path=self.path.split('/')[-1],
                          data=data.strip().decode('utf-8'))

    def errReceived(self, data):
        """Called whenever data is received through stderr"""
        self.log.error(data)

    def processExited(self, status: Failure):
        """Called when process has exited."""

        rc = status.value.exitCode
        if self.log is not None:
            self.log.info("Process ended with exit code {rc}", rc=rc)

        self.deferred.callback(rc)


class PysmurfController:
    """
    Controller object for running pysmurf scripts and functions.

    Args:
        agent (ocs.ocs_agent.OCSAgent):
            OCSAgent object which is running
        args (Namespace):
            argparse namespace with site_config and agent specific arguments

    Attributes:
        agent (ocs.ocs_agent.OCSAgent):
            OCSAgent object which is running
        log (txaio.tx.Logger):
            txaio logger object created by agent
        prot (PysmurfScriptProtocol):
            protocol used to call and monitor external pysmurf scripts
        lock (ocs.ocs_twisted.TimeoutLock):
            lock to protect multiple pysmurf scripts from running simultaneously.
        slot (int):
            ATCA Slot of the smurf-card this agent is commanding.
    """

    def __init__(self, agent, args):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log

        self.prot = None
        self.lock = TimeoutLock()

        self.current_session = None

        self.slot = args.slot
        if self.slot is None:
            self.slot = os.environ['SLOT']

        if args.monitor_id is not None:
            self.agent.subscribe_on_start(
                self._on_session_data,
                'observatory.{}.feeds.pysmurf_session_data'.format(args.monitor_id),
            )

    def _on_session_data(self, _data):
        data, feed = _data

        if self.current_session is not None:
            if data['id'] == os.environ.get("SMURFPUB_ID"):
                if data['type'] == 'session_data':
                    if isinstance(data['payload'], dict):
                        self.current_session.data.update(data['payload'])
                    else:
                        self.log.warn("Session data not passed as a dict!! Skipping...")

                elif data['type'] == 'session_log':
                    if isinstance(data['payload'], str):
                        self.current_session.add_message(data['payload'])

    def _get_smurf_control(self, session=None, load_tune=True, **kwargs):
        """
        Gets pysmurf and det-config instances for sodetlib functions.
        """
        cfg = DetConfig()
        cfg.load_config_files(slot=self.slot)
        S = cfg.get_smurf_control(**kwargs)
        if load_tune:
            S.load_tune(cfg.dev.exp['tunefile'])
        S._ocs_session = session
        return S, cfg

    @inlineCallbacks
    def _run_script(self, script, args, log, session):
        """
        Runs a pysmurf control script. Can only run from the reactor.

        Args:
            script (string):
                path to the script you wish to run
            args (list, optional):
                List of command line arguments to pass to the script.
                Defaults to [].
            log (string or bool, optional):
                Determines if and how the process's stdout should be logged.
                You can pass the path to a logfile, True to use the agent's log,
                or False to not log at all.
        """

        with self.lock.acquire_timeout(0, job=script) as acquired:
            if not acquired:
                return False, "The requested script cannot be run because " \
                              "script {} is already running".format(self.lock.job)

            self.current_session = session
            try:
                # IO  is not really safe from the reactor thread, so we possibly
                # need to find another way to do this if people use it and it
                # causes problems...
                logger = None
                if isinstance(log, str):
                    self.log.info("Logging output to file {}".format(log))
                    log_file = yield threads.deferToThread(open, log, 'a')
                    logger = Logger(observer=FileLogObserver(log_file, log_formatter))
                elif log:
                    # If log==True, use agent's logger
                    logger = self.log

                self.prot = PysmurfScriptProtocol(script, log=logger)
                self.prot.deferred = Deferred()
                python_exec = sys.executable

                cmd = [python_exec, '-u', script] + list(map(str, args))

                self.log.info("{exec}, {cmd}", exec=python_exec, cmd=cmd)

                reactor.spawnProcess(self.prot, python_exec, cmd, env=os.environ)

                rc = yield self.prot.deferred

                return (rc == 0), "Script has finished with exit code {}".format(rc)

            finally:
                # Sleep to allow any remaining messages to be put into the
                # session var
                yield dsleep(1.0)
                self.current_session = None

    @inlineCallbacks
    def run(self, session, params=None):
        """run(script, args=[], log=True)

        **Task** - Runs a pysmurf control script.

        Parameters:
            script (string):
                Path of the pysmurf script to run.
            args (list, optional):
                List of command line arguments to pass to the script.  Defaults
                to [].
            log (string/bool, optional):
                Determines if and how the process's stdout should be logged.
                You can pass the path to a logfile, True to use the agent's
                log, or False to not log at all.

        Notes:
            Data and logs may be passed from the pysmurf control script to the
            session object by publishing it via the Pysmurf Publisher using the
            message types ``session_data`` and ``session_logs`` respectively.

            For example, below is a simple script which starts the data stream
            and returns the datfile path and the list of active channels to the
            session::

                active_channels = S.which_on(0)
                datafile = S.stream_data_on()
                S.pub.publish({
                    'datafile': datafile, 'active_channels': active_channels
                }, msgtype='session_data')

            This would result in the following session.data object::

                >>> response.session['data']
                {
                    'datafile': '/data/smurf_data/20200316/1584401673/outputs/1584402020.dat',
                    'active_channels': [0,1,2,3,4]
                }

        """
        ok, msg = yield self._run_script(
            params['script'],
            params.get('args', []),
            params.get('log', True),
            session
        )

        return ok, msg

    def abort(self, session, params=None):
        """abort()

        **Task** - Aborts the actively running script.

        """
        self.prot.transport.signalProcess('KILL')
        return True, "Aborting process"

    @ocs_agent.param('poll_interval', type=float, default=10)
    @ocs_agent.param('test_mode', default=False, type=bool)
    def check_state(self, session, params=None):
        """check_state(poll_interval=10)

        **Process** - Continuously checks the current state of the smurf. This
        will not modify the smurf state, so this task can be run in conjunction
        with other smurf operations. This will continuously poll smurf metadata
        and update the ``session.data`` object.

        Args
        -----
        poll_interval : float
            Time (sec) between updates.

        Notes
        -------
        The following data will be written to the session.data object::

            >> response.session['data']
            {
                'channel_mask': Array of channels that are streaming data,
                'downsample_factor': downsample_factor,
                'agg_time': Buffer time per G3Frame (sec),
                'open_g3stream': True if data is currently streaming to G3,
                'pysmurf_action': Current pysmurf action,
                'pysmurf_action_timestamp': Current pysmurf-action timestamp,
                'stream_tag': stream-tag for the current g3 stream,
                'last_update':  Time that session-data was last updated,
                'stream_id': Stream-id of the controlled smurf instance
            }
        """
        S, cfg = self._get_smurf_control(load_tune=False, no_dir=True)
        reg = sdl.Registers(S)

        session.set_status('running')
        kw = {'retry_on_fail': False}
        while session.status in ['starting', 'running']:
            try:
                d = dict(
                    channel_mask=S.get_channel_mask(**kw).tolist(),
                    downsample_factor=S.get_downsample_factor(**kw),
                    agg_time=reg.agg_time.get(**kw),
                    open_g3stream=reg.open_g3stream.get(**kw),
                    pysmurf_action=reg.pysmurf_action.get(**kw, as_string=True),
                    pysmurf_action_timestamp=reg.pysmurf_action_timestamp.get(**kw),
                    stream_tag=reg.stream_tag.get(**kw, as_string=True),
                    last_update=time.time(),
                    stream_id=cfg.stream_id,
                )
                session.data.update(d)
            except RuntimeError:
                self.log.warn("Could not connect to epics server! Waiting and "
                              "then trying again")

            time.sleep(params['poll_interval'])

            if params['test_mode']:
                break

        return True, "Finished checking state"

    def _stop_check_state(self, session, params):
        """Stopper for check state process"""
        session.set_status('stopping')

    @ocs_agent.param("duration", default=None, type=float)
    @ocs_agent.param('kwargs', default=None)
    @ocs_agent.param('load_tune', default=True, type=bool)
    @ocs_agent.param('test_mode', default=False, type=bool)
    def stream(self, session, params):
        """stream(duration=None)

        **Process** - Stream smurf data. If a duration is specified, stream
        will end after that amount of time. If unspecified, the stream will run
        until the stop function is called.

        Args
        -----
        duration : float, optional
            If set, determines how many seconds to stream data. By default,
            will leave stream open until stop function is called.
        kwargs : dict
            A dictionary containing additional keyword arguments to pass
            to sodetlib's ``stream_g3_on`` function
        load_tune : bool
            If true, will load a tune-file to the pysmurf object on
            instantiation.

        Notes
        ------
        The following data will be written to the session.data object::

            >> response.session['data']
            {
                'stream_id': Stream-id for the slot,
                'sid': Session-id for the streaming session,
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='stream') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            S, cfg = self._get_smurf_control(session=session,
                                             load_tune=params['load_tune'])

            stop_time = None
            if params['duration'] is not None:
                stop_time = time.time() + params['duration']

            session.data['stream_id'] = cfg.stream_id
            session.data['sid'] = sdl.stream_g3_on(S, **params['kwargs'])
            session.set_status('running')
            while session.status in ['starting', 'running']:
                if stop_time is not None:
                    if time.time() > stop_time:
                        break
                time.sleep(1)

                if params['test_mode']:
                    break
            sdl.stream_g3_off(S)

        return True, 'Finished streaming data'

    def _stream_stop(self, session, params=None):
        session.set_status('stopping')
        return True, "Requesting to end stream"

    @ocs_agent.param('bands', default=None)
    @ocs_agent.param('kwargs', default=None)
    def uxm_setup(self, session, params):
        """uxm_setup(bands=None, kwargs=None)

        **Task** - Runs first-time setup procedure for a UXM. This will run the
        following operations:

            1. Setup Amps (~1 min)
            2. Estimate attens if attens are not already set in the device cfg
               (~1 min / band)
            3. Estimate phase delay (~1 min / band)
            4. Setup tune (~7 min / band)
            5. Setup tracking param (~30s / band)
            6. Measure noise (~45s)

        See the `sodetlib setup docs
        <https://simons1.princeton.edu/docs/sodetlib/operations/setup.html#first-time-setup>`_
        for more information on the sodetlib setup procedure and allowed
        keyword arguments.

        Args
        -----
        bands : list, int
            Bands to set up. Defaults to all.
        kwargs : dict
            Dict containing additional keyword args to pass to the uxm_setup
            function.

        Notes
        -------
        SODETLIB functions such as ``uxm_setup`` and any other functions called
        by ``uxm_setup`` will add relevant data to the ``session.data`` object
        to  a unique key. For example, if all is successful ``session.data``
        may look like::

            >> response.session['data']
            {
                'timestamps': [('setup_amps', 1651162263.0204525), ...],
                'setup_amps_summary': {
                   'success': True,
                   'amp_50k_Id': 15.0,
                   'amp_hemt_Id': 8.0,
                   'amp_50k_Vg': -0.52,
                   'amp_hemt_Vg': -0.829,
                },
                'setup_phase_delay': {
                    'bands': [0, 1, ...]
                    'band_delay_us': List of band delays
                },
                'noise': {
                   'band_medians': List of median white noise for each band
                }
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='uxm_setup') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            S, cfg = self._get_smurf_control(session=session)
            session.set_status('running')
            self.log.info("Starting UXM setup")
            success, summary = uxm_setup.uxm_setup(
                S, cfg, bands=params['bands'], **params['kwargs']
            )

            if not success:
                final_step = session.data['timestamps'][-1][0]
                return False, f"Failed on step {final_step}"
            else:
                return True, "Completed full UXM-Setup procedure"

    @ocs_agent.param('bands', default=None)
    @ocs_agent.param('kwargs', default=None)
    def uxm_relock(self, session, params):
        """uxm_relock(bands=None, kwargs=None)

        **Task** - Relocks detectors to existing tune if setup has already been
        run. Runs the following operations:

            1. Setup Amps (~1 min)
            2. Relocks detectors, setup notches (if requested), and serial
               gradient descent / eta scan (~5 min / band)
            3. Tracking setup (~20s / band)
            4. Noise check (~45s)

        See the `sodetlib relock docs
        <https://simons1.princeton.edu/docs/sodetlib/operations/setup.html#relocking>`_
        for more information on the sodetlib relock procedure and allowed
        keyword arguments.

        Args
        -----
        bands : list, int
            Bands to set up. Defaults to all.
        kwargs : dict
            Dict containing additional keyword args to pass to the uxm_relock
            function.

        Notes
        -------
        SODETLIB functions such as ``uxm_relock`` and any other functions called
        will add relevant data to the ``session.data`` object to  a unique key.
        For example, if all is successful ``session.data`` may look like::

            >> response.session['data']
            {
                'timestamps': [('setup_amps', 1651162263.0204525), ...],
                'setup_amps_summary': {
                   'success': True,
                   'amp_50k_Id': 15.0,
                   'amp_hemt_Id': 8.0,
                   'amp_50k_Vg': -0.52,
                   'amp_hemt_Vg': -0.829,
                },
                'noise': {
                   'band_medians': List of median white noise for each band
                }
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='uxm_relock') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('running')
            S, cfg = self._get_smurf_control(session=session)
            success, summary = uxm_relock.uxm_relock(
                S, cfg, bands=params['bands'], **params['kwargs']
            )
            return success, "Finished UXM Relock"

    @ocs_agent.param('duration', default=30., type=float)
    @ocs_agent.param('kwargs', default=None)
    def take_noise(self, session, params):
        """take_noise(duration=30., kwargs=None)

        **Task** - Takes a short timestream and calculates noise statistics.
        Median white noise level for each band will be stored in the session
        data. See the `sodetlib noise docs
        <https://simons1.princeton.edu/docs/sodetlib/noise.html>`_ for more
        information on the noise function and possible keyword arguments.

        Args
        -----
        duration : float
            Duration of timestream to take for noise calculation.
        kwargs : dict
            Dict containing additional keyword args to pass to the take_noise
            function.

        Notes
        -------
        Median white noise levels for each band will be stored in the
        session.data object, for example::

            >> response.session['data']
            {
                'noise': {
                   'band_medians': List of median white noise for each band
                }
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='take_noise') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('running')
            S, cfg = self._get_smurf_control(session=session)
            sdl.noise.take_noise(S, cfg, params['duration'], **params['kwargs'])
            return True, "Finished taking noise"

    @ocs_agent.param('kwargs', default=None)
    def take_bgmap(self, session, params):
        """take_bgmap(kwargs=None)

        **Task** - Takes a bias-group map. This will calculate the number of
        channels assigned to each bias group and put that into the session data
        object along with the filepath to the analyzed bias-step output. See
        the `bias steps docs page <https://simons1.princeton.edu/docs/sodetlib/operations/bias_steps.html>`_
        for more information on what additional keyword arguments can be
        passed.

        Args
        ----
        kwargs : dict
            Additional kwargs to pass to take_bgmap function.

        Notes
        ------
        The filepath of the BiasStepAnalysis object and the number of channels
        assigned to each bias group will be written to the session.data
        object::

            >> response.session['data']
            {
                'nchans_per_bg': [123, 183, 0, 87, ...],
                'filepath': /path/to/bias_step_file/on/smurf_server.npy,
            }

        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='take_bgmap') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('running')
            S, cfg = self._get_smurf_control(session=session)
            bsa = bias_steps.take_bgmap(S, cfg, **params['kwargs'])
            nchans_per_bg = [0 for _ in range(NBIASLINES + 1)]
            for bg in range(NBIASLINES):
                nchans_per_bg[bg] = int(np.sum(bsa.bgmap == bg))
            nchans_per_bg[-1] = int(np.sum(bsa.bgmap == -1))

            session.data = {
                'nchans_per_bg': nchans_per_bg,
                'filepath': bsa.filepath,
            }
            return True, "Finished taking bgmap"

    @ocs_agent.param('kwargs', default=None)
    def take_iv(self, session, params):
        """take_iv(kwargs=None)

        **Task** - Takes an IV. This will add the normal resistance array and
        channel info to the session data object along with the analyzed IV
        filepath. See the `sodetlib IV docs page
        <https://simons1.princeton.edu/docs/sodetlib/operations/iv.html>`_
        for more information on what additional keyword arguments can be passed
        in.

        Args
        ----
        kwargs : dict
            Additional kwargs to pass to the ``take_iv`` function.

        Notes
        ------
        The following data will be written to the session.data object::

            >> response.session['data']
            {
                'bands': Bands number of each resonator
                'channels': Channel number of each resonator
                'bgmap': BGMap assignment for each resonator
                'R_n': Normal resistance for each resonator
                'filepath': Filepath of saved IVAnalysis object
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='take_iv') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('starting')
            S, cfg = self._get_smurf_control(session=session)
            iva = iv.take_iv(S, cfg, **params['kwargs'])
            session.data = {
                'bands': iva.bands.tolist(),
                'channels': iva.channels.tolist(),
                'bgmap': iva.bgmap.tolist(),
                'R_n': iva.R_n.tolist(),
                'filepath': iva.filepath,
            }
            return True, "Finished taking IV"

    @ocs_agent.param('kwargs', default=None)
    def take_bias_steps(self, session, params):
        """take_bias_steps(kwargs=None)

        **Task** - Takes bias_steps and saves the output filepath to the
        session data object. See the `sodetlib bias step docs page
        <https://simons1.princeton.edu/docs/sodetlib/operations/bias_steps.html>`_
        for more information on bias steps and what kwargs can be passed in.

        Args
        ----
        kwargs : dict
            Additional kwargs to pass to ``take_bais_steps`` function.

        Notes
        ------
        The following data will be written to the session.data object::

            >> response.session['data']
            {
                'filepath': Filepath of saved BiasStepAnalysis object
            }
        """

        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='bias_steps') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('starting')
            S, cfg = self._get_smurf_control(session=session)
            bsa = bias_steps.take_bias_steps(
                S, cfg, **params['kwargs']
            )
            session.data = {
                'filepath': bsa.filepath
            }

            return True, "Finished taking bias steps"

    @ocs_agent.param('rfrac', default=(0.3, 0.6))
    @ocs_agent.param('kwargs', default=None)
    def bias_dets(self, session, params):
        """bias_dets(rfrac=(0.3, 0.6), kwargs=None)

        **Task** - Biases detectors to a target Rfrac value or range. This
        function uses IV results to determine voltages for each bias-group. If
        rfrac is set to be a value, the bias voltage will be set such that the
        median rfrac across all channels is as close as possible to the set
        value. If a range is specified, the voltage will be chosen to maximize
        the number of channels in that range.

        See the sodetlib docs page for `biasing dets into transition
        <https://simons1.princeton.edu/docs/sodetlib/operations/iv.html#biasing-detectors-into-transition>`_
        for more information on the functions and additional keyword args that
        can be passed in.

        Args
        -------
        rfrac : float, tuple
            Target rfrac range to aim for. If this is a float, bias voltages
            will be chosen to get the median rfrac of each bias group as close
            as possible to that value. If
        kwargs : dict
            Additional kwargs to pass to the ``bias_dets`` function.

        Notes
        ------
        The following data will be written to the session.data object::

            >> response.session['data']
            {
                'biases': List of voltage bias values for each bias-group
            }
        """
        if params['kwargs'] is None:
            params['kwargs'] = {}

        with self.lock.acquire_timeout(0, job='bias_steps') as acquired:
            if not acquired:
                return False, f"Operation failed: {self.lock.job} is running."

            session.set_status('running')
            S, cfg = self._get_smurf_control(session=session)
            if isinstance(params['rfrac'], (int, float)):
                biases = bias_dets.bias_to_rfrac(
                    S, cfg, rfrac=params['rfrac'], **params['kwargs']
                )
            else:
                biases = bias_dets.bias_to_rfrac_range(
                    S, cfg, rfrac_range=params['rfrac'], **params['kwargs']
                )

            session.data['biases'] = biases.tolist()

            return True, "Finished taking bias steps"


def make_parser(parser=None):
    """
    Builds argsparse parser, allowing sphinx to auto-document it.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--monitor-id', '-m', type=str,
                        help="Instance id for pysmurf-monitor corresponding to "
                             "this pysmurf instance.")
    pgroup.add_argument('--slot', type=int,
                        help="Smurf slot that this agent will be controlling")
    pgroup.add_argument('--poll-interval', type=float,
                        help="Time between check-state polls")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='PysmurfController',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    controller = PysmurfController(agent, args)

    agent.register_task('run', controller.run, blocking=False)
    agent.register_task('abort', controller.abort, blocking=False)

    startup_pars = {}
    if args.poll_interval is not None:
        startup_pars['poll_interval'] = args.poll_interval
    agent.register_process(
        'check_state', controller.check_state, controller._stop_check_state,
        startup=startup_pars
    )
    agent.register_process(
        'stream', controller.stream, controller._stream_stop
    )
    agent.register_task('uxm_setup', controller.uxm_setup)
    agent.register_task('uxm_relock', controller.uxm_relock)
    agent.register_task('take_bgmap', controller.take_bgmap)
    agent.register_task('take_iv', controller.take_iv)
    agent.register_task('take_bias_steps', controller.take_bias_steps)
    agent.register_task('take_noise', controller.take_noise)
    agent.register_task('bias_dets', controller.bias_dets)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Author: zatkins, zhuber
# Acknowledgments: LS372 agent -- bkoopman, mhasselfield, jlashner

import argparse
import time

import numpy as np
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.Lakeshore.Lakeshore336 import LS336


class LS336_Agent:
    """Agent to connect to a single Lakeshore 336 device.
    Supports channels 'A','B','C', and 'D' for Lakeshore 336s that
    do not have the extra Lakeshore 3062 scanner installed. Also has
    channels 'D2','D3','D4', and 'D5' for 336s that have the extra
    scanner. Currently only supports heaters '1' and '2'.

    Parameters
    ----------

    sn: str
        Serial number of the LS336
    ip: str
        IP Address for the 336 device
    f_sample: float, optional (default 0.1)
        The frequency of sampling for acquiring data (in Hz)
    threshold: float, optional (default 0.1)
        The max difference (in K) between the setpoint and current
        temperature that will be considered stable
    window: int, optional (default 900)
        The amount of time (in s) over which the difference between the
        setpoint and the current temperature must not exceed threshold
        while checking for temperature stability.

    Attributes
    ----------
    sn: str
        Serial number of the LS336
    ip: str
        IP Address for the 336 device
    module: LS336 object
        Driver object
    module.channels: dict
        The available channels in the LS336 object
    module.heaters: dict
        The available heaters in the LS336 object
    f_sample: float
        The frequency of sampling for acquiring data (in Hz)
    t_sample: float
        The time between each sample (inverse of f_sample - 0.01)
    threshold: float
        The max difference (in K) between the setpoint and current temperature
        that will be considered stable
    window: int
        The amount of time (in s) over which the difference between the
        setpoint and the current temperature must not exceed threshold
        while checking for temperature stability.
    _recent_temps: numpy array, protected
        Array of recent temperatures for checking temperature stability
    _static_setpoint: float, protected
        The final setpoint value to avoid issues when the setpoint is
        ramping to a new value. Used in checking temperature stability
    """

    def __init__(self, agent, sn, ip, f_sample=0.1,
                 threshold=0.1, window=900):
        self.agent = agent
        self.sn = sn
        self.ip = ip
        self.f_sample = f_sample
        self.t_sample = 1 / self.f_sample - 0.01
        assert self.t_sample < 7200, \
            "acq sampling freq must be such that t_sample is less than 2 hours"

        self._lock = TimeoutLock()
        self.log = agent.log
        self.initialized = False
        self.take_data = False

        self.module = None

        # for stability checking
        self.threshold = threshold
        self.window = window
        self._recent_temps = None
        self._static_setpoint = None

        agg_params = {'frame_length': 10 * 60}  # sec

        # combined feed for thermometry and control data
        self.agent.register_feed(
            'temperatures',
            record=True,
            agg_params=agg_params,
            buffer_time=1
        )

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init_lakeshore(self, session, params=None):
        """init_lakeshore(auto_acquire=False)

        **Task** - Perform first time setup of the Lakeshore 336 communication

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.
        """
        if params is None:
            params = {}

        # test if this agent is already running
        if self.initialized:
            self.log.info('Lakeshore already initialized, returning...')
            return True, 'Already initialized'

        # initialize lakeshore
        with self._lock.acquire_timeout(job='init', timeout=0) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get lakeshore
            self.module = LS336(self.ip)
            session.add_message(
                f'Lakeshore initialized with ID: {self.module.id}')

        self.initialized = True

        # start data acq if passed
        if params.get('auto_acquire', False):
            self.agent.start('acq')

        return True, 'Lakeshore module initialized'

    @ocs_agent.param('f_sample', default=0.1, type=float)
    def acq(self, session, params=None):
        """acq(f_sample=0.1)

        **Process** - Begins recording data from thermometers and heaters.

        Parameters:
            f_sample (float, optional): Default is 0.1. Sets the
                sampling rate in Hz.

        Notes:
            The most recent data collected is stored in session.data in the
            structure:

               >>> response.session['data']
               {"ls336_fields":
                   {"timestamp": 1921920543,
                    "block_name": "temperatures"
                    "data": {"Channel_A_T": (some value)
                             "Channel_A_V": (some value)
                             "Channel_B_T": (some value)
                             "Channel_B_V": (some value)
                             "Channel_C_T": (some value)
                             "Channel_C_V": (some value)
                             "Channel_D_T": (some value)
                             "Channel_D_V": (some value)
                            }
                   }
               }
        """
        if params is None:
            params = {}

        # get sampling frequency
        f_sample = params.get('f_sample')
        if f_sample is None:
            t_sample = self.t_sample
        else:
            t_sample = 1 / f_sample - 0.01
            self.t_sample = t_sample

        # acquire lock and start Process
        with self._lock.acquire_timeout(job='acq', timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # initialize recent temps array
            # shape is N_points x N_channels
            # N_points is 2 hour / t_sample rounded up
            # N_channels is 8 if the extra scanner is installed, 4 otherwise
            # t_sample can't be more than 2 hours
            N_channels = len(self.module.channels)
            self._recent_temps = np.full(
                (int(np.ceil(7200 / self.t_sample)), N_channels), -1.0)
            print(self._recent_temps.size)

            # acquire data from Lakeshore
            self.take_data = True
            while self.take_data:

                # get thermometry data
                current_time = time.time()
                temperatures_message = {
                    'timestamp': current_time,
                    'block_name': 'temperatures',
                    'data': {}
                }

                temps = self.module.get_kelvin('0')  # array of 4 (or 8) floats
                voltages = self.module.get_sensor('0')  # array of 4/8 floats
                for i, channel in enumerate(self.module.channels.values()):
                    channel_str = channel.input_name.replace(' ', '_')
                    temperatures_message['data'][channel_str + '_T'] = temps[i]
                    temperatures_message['data'][channel_str
                                                 + '_V'] = voltages[i]

                # append to recent temps array for temp stability check
                self._recent_temps = np.roll(self._recent_temps, 1, axis=0)
                self._recent_temps[0] = temps

                # publish to feed
                self.agent.publish_to_feed(
                    'temperatures', temperatures_message)

                # For session.data - named to avoid conflicting with LS372
                # if in use at same time.
                session.data['ls336_fields'] = temperatures_message

                # get heater data
                heaters_message = {
                    'timestamp': current_time,
                    'block_name': 'heaters',
                    'data': {}
                }

                for i, heater in enumerate(self.module.heaters.values()):
                    heater_str = heater.output_name.replace(' ', '_')
                    heaters_message['data'][
                        heater_str + '_Percent'] = heater.get_heater_percent()
                    heaters_message['data'][
                        heater_str + '_Range'] = heater.get_heater_range()
                    heaters_message['data'][
                        heater_str + '_Max_Current'] = heater.get_max_current()
                    heaters_message['data'][
                        heater_str + '_Setpoint'] = heater.get_setpoint()

                # publish to feed
                self.agent.publish_to_feed('temperatures', heaters_message)

                # finish sample
                self.log.debug(
                    f'Sleeping for {np.round(self.t_sample)} seconds...')

                # release and reacquire lock between data acquisition
                self._lock.release()
                time.sleep(t_sample)
                if not self._lock.acquire(timeout=10, job='acq'):
                    print(
                        f"Lock could not be acquired because it is held by "
                        f"{self._lock.job}")
                    return False, 'Could not re-acquire lock'

        return True, 'Acquisition exited cleanly'

    @ocs_agent.param('_')
    def stop_acq(self, session, params=None):
        """stop_acq()

        **Task** - Stops acq process.

        """
        if params is None:
            params = {}

        if self.take_data:
            self.take_data = False
            return True, 'Requested to stop taking data'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('range', type=str,
                     choices=['off', 'low', 'medium', 'high'])
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_heater_range(self, session, params):
        """set_heater_range(range=None,heater='2')

        **Task** - Adjusts the heater range for servoing the load.

        Parameters:
            range (str): Sets the range of the chosen heater. Must be one of
                         'off', 'low', 'medium', and 'high'.
            heater (str, optional): default '2'. Chooses which heater's range
                                    to change. Must be '1' or '2'.

        Notes:
            The range setting has no effect if an output is in the Off mode,
            and it does not apply to an output in Monitor Out mode. An output
            in Monitor Out mode is always on.
        """
        with self._lock.acquire_timeout(job='set_heater_range',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set range
            current_range = heater.get_heater_range()
            if params['range'] == current_range:
                print(
                    'Current heater range matches commanded value. '
                    'Proceeding unchanged')
            else:
                heater.set_heater_range(params['range'])

            session.add_message(
                f"Set {heater.output_name} range to {params['range']}")

        return True, f"Set {heater.output_name} range to {params['range']}"

    @ocs_agent.param('P', type=float, check=lambda x: 0.1 <= x <= 1000)
    @ocs_agent.param('I', type=float, check=lambda x: 0.1 <= x <= 1000)
    @ocs_agent.param('D', type=float, check=lambda x: 0 <= x <= 200)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_pid(self, session, params):
        """set_pid(P=None,I=None,D=None,heater='2')

        **Task** - Set the PID parameters for servoing the load.

        Parameters:
            P (float): Proportional term for PID loop (must be between
                       0.1 and 1000)
            I (float): Integral term for PID loop (must be between 0.1
                       and 1000)
            D (float): Derivative term for PID loop (must be between 0 and 200)
            heater (str, optional): Selects the heater on which to change
                                    the PID settings. Must be '1' or '2'.

        """
        with self._lock.acquire_timeout(job='set_pid', timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set pid
            current_p, current_i, current_d = heater.get_pid()
            if (params['P'] == current_p
                    and params['I'] == current_i
                    and params['D'] == current_d):
                print('Current heater PID matches commanded value. '
                      'Proceeding unchanged')
            else:
                heater.set_pid(params['P'], params['I'], params['D'])

            session.add_message(
                f"Set {heater.output_name} PID to {params['P']}, "
                f"{params['I']}, {params['D']}")

        return True, (f"Set {heater.output_name} PID to {params['P']}, "
                      f" {params['I']}, {params['D']}")

    @ocs_agent.param('mode', type=str, choices=['off', 'closed loop', 'zone',
                                                'open loop'])
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_mode(self, session, params):
        """set_mode(mode=None,heater='2')

        **Task** - Sets the output mode of the heater.

        Parameters:
            mode (str): Selects the output mode for the heater.
                        Accepts four options: 'off', 'closed loop', 'zone',
                        and 'open loop'.
                        for restrictions based on the selected heater.
            heater (str, optional): Default '2'. Selects the heater on which
                                    to change the mode. Must be '1' or '2'.

        Notes:
            Does not support the options 'monitor out' and 'warm up',
            which only work for the unsupported analog outputs
            (heaters 3 and 4).
        """
        with self._lock.acquire_timeout(job='set_mode', timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set mode
            current_mode = heater.get_mode()
            if params['mode'] == current_mode:
                print(
                    'Current heater mode matches commanded value. '
                    'Proceeding unchanged')
            else:
                heater.set_mode(params['mode'])

            session.add_message(
                f"Set {heater.output_name} mode to {params['mode']}")

        return True, f"Set {heater.output_name} mode to {params['mode']}"

    @ocs_agent.param('resistance', type=float)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_heater_resistance(self, session, params):
        """set_heater_resistance(resistance=None,heater='2')

        **Task** - Sets the heater resistance and resistance setting
        of the heater. The associated 'get' function in the Heater class
        is get_heater_resistance_setting().

        Parameters:
            resistance (float): The actual resistance of the load
            heater (str, optional): Default '2'. Selects the heater on which
                                    to set the resistance. Must be '1' or '2'.
        """
        with self._lock.acquire_timeout(job='set_heater_resistance',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set heater resistance
            _ = heater.get_heater_resistance_setting()
            if params['resistance'] == heater.resistance:
                print(
                    'Current heater resistance matches commanded value. '
                    'Proceeding unchanged')
            else:
                heater.set_heater_resistance(params['resistance'])

            session.add_message(
                f"Set {heater.output_name} resistance to "
                f"{params['resistance']}")

        return True, (f"Set {heater.output_name} resistance to "
                      f"{params['resistance']}")

    @ocs_agent.param('current', type=float, check=lambda x: 0.0 <= x <= 2.0)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_max_current(self, session, params):
        """set_max_current(current=None,heater='2')

        **Task** - Sets the maximum current that can pass through a heater.

        Parameters:
            current (float): The desired max current. Must be between
                             0 and 2 A.
            heater (str, optional): Default '2'. Selects the heater on which
                                    to set the max current. Must be '1' or '2'.
        """
        with self._lock.acquire_timeout(job='set_max_current',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set max current
            current_max_current = heater.get_max_current()
            if params['current'] == current_max_current:
                print(
                    'Current max current matches commanded value. '
                    'Proceeding unchanged')
            else:
                heater.set_max_current(params['current'])

            session.add_message(
                f"Set {heater.output_name} max current to {params['current']}")

        return True, (f"Set {heater.output_name} max current to "
                      f"{params['current']}")

    @ocs_agent.param('percent', type=float)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_manual_out(self, session, params):
        """set_manual_out(percent=None,heater='2')

        **Task** - Sets the manual output of the heater as a percentage of the
        full current or power depending on which display the heater
        is set to use.

        Parameters:
            percent (float): Percent of full current or power to set on the
                             heater. Must have 2 or fewer decimal places.
            heater (str, optional): Default '2'. Selects the heater on which
                                    to set the manual output.
                                    Must be '1' or '2'.
        """
        with self._lock.acquire_timeout(job='set_manual_out',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set manual out
            current_manual_out = heater.get_manual_out()
            if params['percent'] == current_manual_out:
                print('Current manual out matches commanded value. '
                      'Proceeding unchanged')
            else:
                heater.set_manual_out(params['percent'])

            session.add_message(
                f"Set {heater.output_name} manual out to {params['percent']}")

        return True, (f"Set {heater.output_name} manual out to "
                      f"{params['percent']}")

    @ocs_agent.param('input', type=str,
                     choices=['A', 'B', 'C', 'D', 'D1', 'D2', 'D3', 'D4', 'D5'])
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_input_channel(self, session, params):
        """set_input_channel(input=None,heater='2')

        **Task** - Sets the input channel of the heater control loop.

        Parameters:
            input (str): The name of the heater to use as the input channel.
                         Must be one of 'none','A','B','C', or 'D'.
                         Can also be 'D2','D3','D4', or 'D5' if the extra
                         Lakeshore 3062 Scanner is installed in your LS336.
            heater (str, optional): Default '2'. Selects the heater for which
                                    to set the input channel.
                                    Must be '1' or '2'.

        """
        with self._lock.acquire_timeout(job='set_input_channel',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # D1 is the same as D
            if params['input'] == 'D1':
                params['input'] = 'D'

            # set input channel
            current_input_channel = heater.get_input_channel()
            if params['input'] == current_input_channel:
                print(
                    'Current input channel matches commanded value. '
                    'Proceeding unchanged')
            else:
                heater.set_input_channel(params['input'])

            session.add_message(
                f"Set {heater.output_name} input channel to {params['input']}")

        return True, (f"Set {heater.output_name} input channel to "
                      f"{params['input']}")

    @ocs_agent.param('setpoint', type=float)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def set_setpoint(self, session, params):
        """set_setpoint(setpoint=None,heater='2')

        **Task** - Sets the setpoint of the heater control loop,
        after first turning ramp off. May be a limit to how high the setpoint
        can go based on your system parameters.

        Parameters:
            setpoint (float): The setpoint for the control loop. Units depend
                              on the preferred sensor units (Kelvin, Celsius,
                              or Sensor).
            heater (str, optional): Default '2'. Selects the heater for which
                                    to set the input channel.
                                    Must be '1' or '2'.
        """
        with self._lock.acquire_timeout(job='set_setpoint',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # set setpoint
            current_setpoint = heater.get_setpoint()
            if params['setpoint'] == current_setpoint:
                print('Current setpoint matches commanded value. '
                      'Proceeding unchanged')
            else:
                heater.set_ramp_on_off('off')
                heater.set_setpoint(params['setpoint'])
                # static setpoint used in temp stability check
                # to avoid ramping bug
                self._static_setpoint = params['setpoint']

            session.add_message(
                f"Turned ramp off and set {heater.output_name} setpoint to "
                f"{params['setpoint']}")

        return True, (f"Turned ramp off and set {heater.output_name} "
                      f"setpoint to {params['setpoint']}")

    @ocs_agent.param('T_limit', type=int)
    @ocs_agent.param('channel', type=str, default='A',
                     choices=['A', 'B', 'C', 'D', 'D2', 'D3', 'D4', 'D5'])
    def set_T_limit(self, session, params):
        """set_T_limit(T_limit=None,channel='A')

        **Task** - Sets the temperature limit above which the control
                   output assigned to the selected channel shut off.

        Parameters:
            T_limit (int): The temperature limit in Kelvin. Note that a limit
                           of 0 K turns off this feature for the given channel.
            channel (str, optional): Default 'A'. Selects which channel to use
                                     for controlling the temperature. Options
                                     are 'A','B','C', and 'D'. Can also be
                                     'D2','D3','D4', or 'D5' if the extra
                                     Lakeshore 3062 Scanner is installed in
                                     your LS336.
        """
        with self._lock.acquire_timeout(job='set_T_limit',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get channel
            channel_key = params.get('channel', 'A')  # default to input A
            channel = self.module.channels[channel_key]

            # set T limit
            current_limit = channel.get_T_limit()
            if params['T_limit'] == current_limit:
                print('Current T limit matches commanded value. '
                      'Proceeding unchanged')
            else:
                channel.set_T_limit(params['T_limit'])

            session.add_message(
                f"Set {channel.input_name} T limit to {params['T_limit']}")

        return True, f"Set {channel.input_name} T limit to {params['T_limit']}"

    @ocs_agent.param('temperature', type=float)
    @ocs_agent.param('ramp', default=0.1, type=float)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    @ocs_agent.param('transport', default=False, type=bool)
    @ocs_agent.param('transport_offset', default=0, type=float,
                     check=lambda x: x >= 0.0)
    def servo_to_temperature(self, session, params):
        """servo_to_temperature(temperature=None,ramp=0.1,heater='2',\
                                transport=False,transport_offset=0)

        **Task** - A wrapper for setting the heater setpoint. Performs sanity
        checks on heater configuration before publishing setpoint:

            1. checks control mode of heater (closed loop)
            2. checks units of input channel (kelvin)
            3. resets setpoint to current temperature with ramp off
            4. sets ramp on to specified rate
            5. checks setpoint does not exceed input channel T_limit
            6. sets setpoint to commanded value

        Note that this function does NOT turn on the heater if it is off. You
        must use set_heater_range() to pick a range first.

        Parameters:
            temperature (float): The new setpoint in Kelvin. Make sure there is
                                 is a control input set to the heater and its
                                 units are Kelvin.
            ramp (float, optional): Default 0.1. The rate for how quickly
                                    the setpoint ramps to new value.
                                    Units of K/min.
            heater (str, optional): Default '2'. The heater to use
                                    for servoing. Must be '1' or '2'.
            transport (bool, optional): Default False. See Notes
                                        for description.
            transport_offset (float, optional): Default 0. In Kelvin.
                                                See Notes.

        Notes:
        If param 'transport' is provided and True, the control loop restarts
        when the setpoint is first reached. This is useful for loads with long
        cooling times or time constant to help minimize over/undershoot.

        If param 'transport' is provided and True, and 'transport_offset' is
        provided and positive, and the setpoint is higher than the
        current temperature, then the control loop will restart
        when the setpoint - transport_offset is first reached.
        This is useful to avoid  a "false positive"
        temperature stability check too shortly after transport completes.
        """
        # get sampling frequency
        t_sample = self.t_sample

        with self._lock.acquire_timeout(job='servo_to_temperature',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # get current setpoint
            current_setpoint = heater.get_setpoint()

            # check in correct control mode
            if heater.get_mode() != 'closed loop':
                session.add_message(
                    'Changing control to closed loop mode for servo.')
                heater.set_mode("closed loop")

            # check in correct units
            channel = heater.get_input_channel()
            if channel == 'none':
                return False, (f'{heater.output_name} does not have an '
                               f'input channel assigned')
            if self.module.channels[channel].get_units() != 'kelvin':
                session.add_message(
                    'Setting preferred units to kelvin on '
                    'heater control input.')
                self.module.channels[channel].set_units('kelvin')

            # restart setpoint at current temperature
            current_temp = np.round(float(self.module.get_kelvin(channel)), 4)
            session.add_message(
                f'Turning ramp off and setting setpoint to current '
                f'temperature {current_temp}')
            heater.set_ramp_on_off('off')
            heater.set_setpoint(current_temp)

            # reset ramp settings
            ramp = params.get('ramp', 0.1)
            session.add_message(
                f'Turning ramp on and setting rate to {ramp}K/min')
            heater.set_ramp_on_off('on')
            heater.set_ramp_rate(ramp)

            # make sure not exceeding channel T limit
            T_limit = self.module.channels[channel].get_T_limit()
            if T_limit <= params['temperature']:
                return False, (f"{heater.output_name} control channel "
                               f"{channel} T limit of {T_limit}K is lower "
                               f"than setpoint of {params['temperature']}")

            # set setpoint
            if params['temperature'] == current_setpoint:
                print('Current setpoint matches commanded value. '
                      'Proceeding unchanged')
            else:
                session.add_message(
                    f"Setting {heater.output_name} setpoint to "
                    f"{params['temperature']}")
                heater.set_setpoint(params['temperature'])
                # static setpoint used in temp stability check
                # to avoid pulling the ramping setpoint
                self._static_setpoint = params['temperature']

                # if transport, restart control loop when setpoint
                # first crossed
                if params.get('transport', False):

                    current_range = heater.get_heater_range()
                    starting_sign = np.sign(
                        params['temperature'] - current_temp)
                    transporting = True

                    # if we are raising temp, allow possibility of
                    # stopping transport at a cooler temp
                    T_offset = 0
                    if starting_sign > 0:
                        T_offset = params.get('transport_offset', 0)
                        if T_offset < 0:
                            return False, ('Transport offset temperature '
                                           'cannot be negative')

                    while transporting:
                        current_temp = np.round(
                            self.module.get_kelvin(channel), 4)
                        # check when this flips
                        current_sign = np.sign(
                            params['temperature'] - T_offset - current_temp)

                        # release and reacquire lock between data acquisition
                        self._lock.release()
                        time.sleep(t_sample)
                        if not self._lock.acquire(timeout=10,
                                                  job='servo_to_temperature'):
                            print(
                                f"Lock could not be acquired because it is "
                                f"held by {self._lock.job}")
                            return False, 'Could not re-acquire lock'

                        if current_sign != starting_sign:
                            transporting = False  # update flag

                            # cycle control loop
                            session.add_message(
                                'Transport complete, restarting control '
                                'loop at provided setpoint')
                            heater.set_heater_range('off')
                            # necessary 1 s for prev command to register
                            # in ls336 firmware for some reason
                            time.sleep(1)
                            heater.set_heater_range(current_range)

        return True, (f"Set {heater.output_name} setpoint to "
                      f"{params['temperature']}")

    @ocs_agent.param('threshold', default=0.1, type=float)
    @ocs_agent.param('window', default=900, type=int)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def check_temperature_stability(self, session, params):
        """check_temperature_stability(threshold=0.1,window=900,heater='2')

        **Task** - Assesses whether the load is stable around the setpoint
        to within some threshold.

        Parameters:
            threshold (float, optional): Default 0.1. See notes.
            window (int, optional): Default 900. See notes.
            heater (str, optional): Default '2'. Selects the heater for which
                                    to set the input channel.
                                    Must be '1' or '2'.

        Notes
        -----
        Param 'threshold' sets the upper bound on the absolute
        temperature difference between the setpoint and any temperature
        from the input channel in the last 'window' seconds.

        Param 'window' sets the lookback time into the most recent
        temperature data, in seconds. Note that this function grabs the most
        recent data in one window-length of time; it does not take new data.

        If you want to use the result of this task for making logical decisions
        in a client (e.g. waiting longer before starting a process if the
        temperature is not yet stable), use the session['success'] key. It will
        be True if the temperature is stable and False if not.
        Example:
        >>> response = ls336.check_temperature_stability()
        >>> response.session['success']
        True
        """

        # get threshold
        threshold = params.get('threshold')
        if threshold is None:
            threshold = self.threshold

        # get window
        window = params.get('window')
        if window is None:
            window = self.window
        num_idxs = int(np.ceil(window / self.t_sample))

        with self._lock.acquire_timeout(job='check_temperature_stability',
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # get channel
            channel = heater.get_input_channel()
            channel_num = self.module.channels[channel].num

            # get current temp
            current_temp = np.round(self.module.get_kelvin(channel), 4)

            # check if recent temps and current temps are within threshold
            _recent_temps = self._recent_temps[:num_idxs, channel_num - 1]
            _recent_temps = np.concatenate(
                (np.array([current_temp]), _recent_temps))

            # get static setpoint if None
            if self._static_setpoint is None:
                self._static_setpoint = heater.get_setpoint()

            # avoids checking against the ramping setpoint,
            # i.e. want to compare to commanded setpoint not mid-ramp setpoint
            setpoint = self._static_setpoint
            session.add_message(
                f'Maximum absolute difference in recent temps is '
                f'{np.max(np.abs(_recent_temps - setpoint))}K')

            if np.all(np.abs(_recent_temps - setpoint) < threshold):
                session.add_message(
                    f'Recent temps are within {threshold}K of setpoint')
                return True, (f'Servo temperature is stable within '
                              f'{threshold}K of setpoint')

        return False, (f'Servo temperature is not stable within '
                       f'{threshold}K of setpoint')

    @ocs_agent.param('attribute', type=str)
    @ocs_agent.param('channel', type=str, default='A',
                     choices=['A', 'B', 'C', 'D', 'D1', 'D2', 'D3', 'D4', 'D5'])
    def get_channel_attribute(self, session, params):
        """get_channel_attribute(attribute=None,channel='A')

        **Task** - Gets an arbitrary channel attribute and stores it in the
        session.data dict. Attribute must be the name of a method
        in the namespace of the Lakeshore336 Channel class,
        with a leading "get\\_" removed (see example).

        Parameters:
            attribute (str): The name of the channel attribute to get. See the
                             Lakeshore 336 Channel class API for all options.
            channel (str, optional): Default 'A'. Selects which channel for
                                     which to get the attribute. Options
                                     are 'A','B','C', and 'D'. Can also be
                                     'D2','D3','D4', or 'D5' if the extra
                                     Lakeshore 3062 Scanner is installed in
                                     your LS336.

        Example:
        >>> ls.get_channel_attribute(attribute = 'T_limit').session['data']
        {'T_limit': 30.0}
        """
        with self._lock.acquire_timeout(job=f"get_{params['attribute']}",
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get channel
            channel_key = params.get('channel', 'A')  # default to input A
            channel = self.module.channels[channel_key]

            # check that attribute is a valid channel method
            if getattr(channel, f"get_{params['attribute']}",
                       False) is not False:
                query = getattr(channel, f"get_{params['attribute']}")

            # get attribute
            resp = query()
            session.data[params['attribute']] = resp

        return True, (f"Retrieved {channel.input_name} {params['attribute']}, value is {resp}")

    @ocs_agent.param('attribute', type=str)
    @ocs_agent.param('heater', default='2', type=str, choices=['1', '2'])
    def get_heater_attribute(self, session, params):
        """get_heater_attribute(attribute=None,heater='2')

        **Task** - Gets an arbitrary heater attribute and stores it
        in the session.data dict. Attribute must be the name of a method
        in the namespace of the Lakeshore336 Heater class, with a leading
        "get\\_" removed (see example).

        Parameters:
            attribute (str): The name of the channel attribute to get. See the
                             Lakeshore 336 Heater class API for all options.
            heater (str, optional): Default '2'. Selects the heater for which
                                    to get the heater attribute.
                                    Must be '1' or '2'.

        Examples
        --------
        >>> ls.get_heater_attribute(attribute = 'heater_range').session['data']
        {'heater_range': 'off'}
        """
        with self._lock.acquire_timeout(job=f"get_{params['attribute']}",
                                        timeout=3) as acquired:
            if not acquired:
                print(
                    f"Lock could not be acquired because it is held by "
                    f"{self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater_key = params.get('heater', '2')  # default to 50W output
            heater = self.module.heaters[heater_key]

            # check that attribute is a valid heater method
            if getattr(heater, f"get_{params['attribute']}",
                       False) is not False:
                query = getattr(heater, f"get_{params['attribute']}")

            # get attribute
            resp = query()
            session.data[params['attribute']] = resp

        return True, f"Retrieved {heater.output_name} {params['attribute']}, value is {resp}"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to
    automatically build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--serial-number')
    pgroup.add_argument('--ip-address', type=str,
                        help="IP address for the lakeshore")
    pgroup.add_argument('--f-sample', type=float, default=0.1,
                        help='The frequency of data acquisition')
    pgroup.add_argument('--threshold', type=float, default=0.1,
                        help='The upper bound on temperature differences '
                             'for stability check')
    pgroup.add_argument('--window', type=float, default=900.,
                        help='The lookback time on temperature differences '
                             'for stability check')
    pgroup.add_argument('--auto-acquire', type=bool, default=True,
                        help='Automatically start data acquisition on startup')
    return parser


def main(args=None):
    # Create an argument parser
    parser = make_parser()
    args = site_config.parse_args(agent_class='Lakeshore336Agent',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested
    init_params = False
    if args.auto_acquire:
        init_params = {'auto_acquire': True}

    print('I am in charge of device with serial '
          'number: %s' % args.serial_number)

    # Create a session and a runner which communicate over WAMP
    agent, runner = ocs_agent.init_site_agent(args)

    # Pass the new agent session to the agent class
    lake_agent = LS336_Agent(agent, args.serial_number, args.ip_address,
                             args.f_sample, args.threshold, args.window)

    # Register tasks (name, agent_function)
    agent.register_task(
        'init_lakeshore', lake_agent.init_lakeshore, startup=init_params)
    agent.register_task('set_heater_range', lake_agent.set_heater_range)
    agent.register_task('set_heater_resistance',
                        lake_agent.set_heater_resistance)
    agent.register_task('set_input_channel', lake_agent.set_input_channel)
    agent.register_task('set_manual_out', lake_agent.set_manual_out)
    agent.register_task('set_max_current', lake_agent.set_max_current)
    agent.register_task('set_mode', lake_agent.set_mode)
    agent.register_task('set_pid', lake_agent.set_pid)
    agent.register_task('set_T_limit', lake_agent.set_T_limit)
    agent.register_task('set_setpoint', lake_agent.set_setpoint)
    agent.register_task('servo_to_temperature',
                        lake_agent.servo_to_temperature)
    agent.register_task('check_temperature_stability',
                        lake_agent.check_temperature_stability)
    agent.register_task('get_channel_attribute',
                        lake_agent.get_channel_attribute)
    agent.register_task('get_heater_attribute',
                        lake_agent.get_heater_attribute)

    # Register processes (name, agent_start_function, agent_end_function)
    agent.register_process('acq', lake_agent.acq, lake_agent.stop_acq)

    # Run the agent
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import socket
import time

import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock
from twisted.internet import reactor


class FTSAerotechStage:
    """
    Class for connecting to the FTS mirror controller

    Args:
        ip_address: IP address where controller is running
        port: Port the controller is listening on
        timeout: communication timeout
        speed: speed in mm/s, defaults to 25 mm/s if None
        translate: Argument which translates the aerotech stage readout into
                   a standardized value.
        limits: 2-tuple of the max and min FTS central mirror positions to
                prevent the stage from moving out of bounds via software
                limits.

    """

    def __init__(self, ip_address, port, translate=None, limits=None,
                 timeout=10, speed=25):
        self.ip_address = ip_address
        self.port = int(port)

        self.comm = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.comm.connect((self.ip_address, self.port))
        self.comm.settimeout(timeout)

        self.send('ENABLE X\n')  # Send enable command
        data = self.comm.recv(1024)  # Collect and print response

        if data == b'%\n':
            self.initialized = True
        else:
            self.initialized = False

        self.pos = None
        self.speed = speed
        self.speed_code = 'F%i' % self.speed

        self.translate = translate
        self.limits = limits

    def send(self, msg):
        self.comm.sendall(bytes(msg, 'utf-8'))

    def read(self):
        # Controller blocks reply until motion is complete; so if
        # you're putting a timeout in here make sure it's smart/long
        # enough.
        return self.comm.recv(1024)

    def home(self):
        self.send('HOME X\n')
        time.sleep(0.1)
        # block until homing is complete
        return self.read()

    def get_position(self):
        self.send('CMDPOS X\n')
        time.sleep(0.1)
        out = self.read()
        try:
            M, B = self.translate
            self.pos = (float(out[1:]) - B) / M
            return True, self.pos
        except BaseException:
            return False, None

    def move_to(self, position):
        limits = self.limits
        if position < limits[0] or position > limits[1]:
            return False, 'Move out of bounds!'
        M, B = self.translate
        stage_pos = position * M + B
        cmd = ('MOVEABS X%.2f %s\n' % (stage_pos, self.speed_code))
        self.send(cmd)
        out = None
        while out is None:
            try:
                # controller should block reply until move complete.
                time.sleep(0.1)
                out = self.read()
            except TimeoutError:
                continue
        return True, 'Move Complete'

    def close(self):
        self.comm.close()


class FTSAerotechAgent:
    """
    Agent for connecting to the FTS mirror control.

    Args:
        ip_addr: IP address of Motion Controller
        port: Port of Motion Controller
        mode: 'acq': Start data acquisition on initialize
        samp: default sampling frequency in Hz
        config_file: File which contains FTS-specific translate, limits, and
                     speed arguments.

    """

    def __init__(self, agent, ip_addr, port, config_file, mode=None, samp=2):

        self.ip_addr = ip_addr
        self.port = int(port)

        self.stage = None
        self.initialized = False
        self.take_data = False

        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        if mode == 'acq':
            self.auto_acq = True
        else:
            self.auto_acq = False
        self.sampling_frequency = float(samp)

        # register the position feeds
        agg_params = {
            'frame_length': 10 * 60,  # [sec]
        }

        self.agent.register_feed('position',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

        # Load dictionary of specific mirror paramters, since some parameters
        # like limits and translate vary over different FTSes This is loaded
        # from a yaml file, which is assumed to be in the $OCS_CONFIG_DIR
        # directory.
        if config_file is None:
            raise Exception(
                "No config file specified for the FTS mirror config")
        else:
            config_file_path = os.path.join(os.environ['OCS_CONFIG_DIR'],
                                            config_file)
            with open(config_file_path) as stream:
                self.mirror_configs = yaml.safe_load(stream)
                if self.mirror_configs is None:
                    raise Exception("No mirror configs in config file.")
                self.log.info(
                    f"Loaded mirror configs from file {config_file_path}")
                self.translate = self.mirror_configs.pop('translate', None)
                self.limits = self.mirror_configs.pop('limits', None)
                # The other mirror configs (speed, timeout) are optional and
                # have defaults so we leave them as the dictionary.
                if self.translate is None or self.limits is None:
                    raise Exception("translate and limits must be included "
                                    "in the mirror configuration keys")

    @ocs_agent.param('_')
    def init_stage(self, session, params=None):
        """init_stage()

        **Task** - Perform first time setup for communication with FTS stage.

        """
        if self.stage is not None and self.initialized:
            return True, 'Stages already Initialized'

        self.log.debug("Trying to acquire lock")
        with self.lock.acquire_timeout(timeout=0, job='init') as acquired:
            # Locking mechanism stops code from proceeding if no lock acquired
            if not acquired:
                self.log.warn("Could not start init because {} is already"
                              "running".format(self.lock.job))
                return False, "Could not acquire lock."
            # Run the function you want to run
            self.log.debug("Lock Acquired Connecting to Stages")
            try:
                self.stage = FTSAerotechStage(
                    self.ip_addr, self.port, self.translate, self.limits,
                    **self.mirror_configs)
            except Exception as e:
                self.log.error(f"Error while connecting to FTS: {e}")
                reactor.callFromThread(reactor.stop)
                return False, "FTS Stage Initialization Failed"
        # This part is for the record and to allow future calls to proceed,
        # so does not require the lock
        self.initialized = True
        if self.auto_acq:
            self.agent.start('acq')
        return True, 'Stage Initialized.'

    @ocs_agent.param('_')
    def home(self, session, params=None):
        """home()

        **Task** - Home the stage to its negative limit.

        """

        with self.lock.acquire_timeout(timeout=3, job='home') as acquired:
            if not acquired:
                self.log.warn("Could not start home because lock held by"
                              f"{self.lock.job}")
                return False, "Could not get lock"
            try:
                self.stage.home()
            except Exception as e:
                self.log.error(f"Homing Failed: {e}")
                return False, "Homing Failed"
        return True, "Homing Complete"

    @ocs_agent.param('position', type=float, check=lambda x: -74.8 <= x <= 74.8)
    def move_to(self, session, params=None):
        """move_to(position)

        **Task** - Move to absolute position relative to stage center (in mm).

        Parameters:
            position (float): Position in mm, must be between -74.8 and 74.8.

        """
        with self.lock.acquire_timeout(timeout=3, job='move') as acquired:
            if not acquired:
                self.log.warn("Could not start move because lock held by"
                              f"{self.lock.job}")
                return False, "Could not get lock"
            return self.stage.move_to(params.get('position'))

        return False, "Move did not complete correctly?"

    @ocs_agent.param('sampling_frequency', type=float, default=2)
    def acq(self, session, params=None):
        """acq(sampling_frequency=2)

        Parameters:
            sampling_frequency (float): Sampling rate in Hz. Defaults to 2 Hz.

        Notes:
            The most recent position data is stored in session.data in the
            format::

                >>> response.session['data']
                {"position": {"pos" : mirror position}}

        """
        f_sample = params.get('sampling_frequency', self.sampling_frequency)
        pm = Pacemaker(f_sample, quantize=True)

        if not self.initialized or self.stage is None:
            raise Exception("Connection to Stages not initialized")

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(f"Could not start acq because {self.lock.job} "
                              "is already running")
                return False, "Could not acquire lock."

            self.log.info("Starting Data Acquisition for FTS Mirror at"
                          f"{f_sample} Hz")
            session.set_status('running')
            self.take_data = True
            last_release = time.time()

            while self.take_data:
                if time.time() - last_release > 1.:
                    if not self.lock.release_and_acquire(timeout=20):
                        self.log.warn("Could not re-acquire lock now held by"
                                      f"{self.lock.job}.")
                        return False, "could not re-acquire lock"
                    last_release = time.time()
                pm.sleep()

                data = {
                    'timestamp': time.time(),
                    'block_name': 'position',
                    'data': {}}
                success, pos = self.stage.get_position()
                if not success:
                    self.log.info("stage.get_position call failed")
                else:
                    data['data']['pos'] = pos
                    self.agent.publish_to_feed('position', data)

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        params:
            dict: {}
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running.'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--port')
    pgroup.add_argument('--config-file')
    pgroup.add_argument('--mode')
    pgroup.add_argument('--sampling_frequency')
    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class='FTSAerotechAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    fts_agent = FTSAerotechAgent(agent, args.ip_address, args.port,
                                 args.config_file, args.mode,
                                 args.sampling_frequency)

    agent.register_task('init_stage', fts_agent.init_stage)
    agent.register_task('move_to', fts_agent.move_to)
    agent.register_task('home', fts_agent.home)

    agent.register_process('acq', fts_agent.acq, fts_agent._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# Script to log and readout pfeiffer TPG 366 gauge contoller
# via Ethernet connection
# Zhilei Xu, Tanay Bhandarkar

import argparse
import socket
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

BUFF_SIZE = 128
ENQ = '\x05'


class Pfeiffer:
    """CLASS to control and retrieve data from the pfeiffer tpg366
    pressure gauge controller


    Args:
        ip_address: IP address of the deivce
        porti (int): 8000 (fixed for the device)

    Attributes:
       read_pressure reads the pressure from one channel (given as an argument)
       read_pressure_all reads pressures from the six channels
       close closes the socket
    """

    def __init__(self, ip_address, port, timeout=10,
                 f_sample=2.5):
        self.ip_address = ip_address
        self.port = port
        self.comm = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.comm.connect((self.ip_address, self.port))
        self.comm.settimeout(timeout)

    def read_pressure(self, ch_no):
        """
        Function to measure the pressure of one given channel
        ch_no is the chanel to be measured (e.g. 1-6)
        returns the measured pressure as a float

        Args:
            ch_no: The channel to be measured (1-6)

        Returns:
            pressure as a float
        """
        msg = 'PR%d\r\n' % ch_no
        self.comm.send(msg.encode())
        # Can use this to catch exemptions, for troubleshooting
        self.comm.recv(BUFF_SIZE).decode()
        self.comm.send(ENQ.encode())
        read_str = self.comm.recv(BUFF_SIZE).decode()
        pressure_str = read_str.split(',')[-1].split('\r')[0]
        pressure = float(pressure_str)
        return pressure

    def read_pressure_all(self):
        """measure the pressure of all channel
        Return an array of 6 pressure values as a float array

        Args:
            None

        Returns:
            6 element array corresponding to each channels
            pressure reading, as floats
        """
        msg = 'PRX\r\n'
        self.comm.send(msg.encode())
        # Could use this to catch exemptions, for troubleshooting
        self.comm.recv(BUFF_SIZE).decode()
        self.comm.send(ENQ.encode())
        read_str = self.comm.recv(BUFF_SIZE).decode()
        pressure_str = read_str.split('\r')[0]
        # gauge_states = pressure_str.split(',')[::2]
        pressures = pressure_str.split(',')[1::2]
        pressures = [float(p) for p in pressures]
        return pressures

    def close(self):
        """Close the socket of the connection"""
        self.comm.close()


class PfeifferAgent:

    def __init__(self, agent, ip_address, port, f_sample=2.5):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.f_sample = f_sample
        self.take_data = False
        self.gauge = Pfeiffer(ip_address, int(port))
        agg_params = {'frame_length': 60, }

        self.agent.register_feed('pressures',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('sampling_frequency', type=float, default=2.5)
    @ocs_agent.param('test_mode', type=bool, default=False)
    def acq(self, session, params=None):
        """acq(sampling_frequency=2.5, test_mode=False)

        **Process** - Get pressures from the Pfeiffer gauges.

        Parameters:
            sampling_frequency (float): Rate at which to get the pressures
                [Hz]. Defaults to 2.5 Hz.
            test_mode (bool): Run the Process loop only once. This is meant
                only for testing. Defaults to False.

        """
        f_sample = params['sampling_frequency']
        if f_sample is None:
            f_sample = self.f_sample

        sleep_time = 1. / f_sample - 0.01

        with self.lock.acquire_timeout(timeout=0, job='init') as acquired:
            # Locking mechanism stops code from proceeding if no lock acquired
            if not acquired:
                self.log.warn("Could not start init because {} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            self.take_data = True

            while self.take_data:
                data = {
                    'timestamp': time.time(),
                    'block_name': 'pressures',
                    'data': {}
                }
                pressure_array = self.gauge.read_pressure_all()
                # Loop through all the channels on the device
                for channel in range(len(pressure_array)):
                    data['data']["pressure_ch" + str(channel + 1)] = pressure_array[channel]

                self.agent.publish_to_feed('pressures', data)
                time.sleep(sleep_time)

                if params['test_mode']:
                    break

            self.agent.feeds['pressures'].flush_buffer()
        return True, 'Acquistion exited cleanly'

    def _stop_acq(self, session, params=None):
        """
        End pressure data acquisition
        """
        if self.take_data:
            self.take_data = False
            self.gauge.close()
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip_address')
    pgroup.add_argument('--port')
    pgroup.add_argument("--mode", type=str, default='acq', choices=['acq', 'test'])

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='PfeifferAgent',
                                  parser=parser,
                                  args=args)

    init_params = True
    if args.mode == 'test':
        init_params = {'test_mode': True}

    agent, runner = ocs_agent.init_site_agent(args)
    pfeiffer_agent = PfeifferAgent(agent, args.ip_address, args.port)
    agent.register_process('acq', pfeiffer_agent.acq,
                           pfeiffer_agent._stop_acq, startup=init_params)
    agent.register_task('close', pfeiffer_agent._stop_acq)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# Script to log and readout PTC data through ethernet connection.
# Tamar Ervin and Jake Spisak, February 2019
# Sanah Bhimani, May 2022

import argparse
import random
import socket
import struct
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

STX = '\x02'
ADDR = '\x10'
CMD = '\x80'
CR = '\x0D'
DATA_WRITE = '\x61'
DATA_READ = '\x63'
ESC = '\x07'
ESC_STX = '\x30'
ESC_CR = '\x31'
ESC_ESC = '\x32'


class PTC:
    def __init__(self, ip_address, port=502, timeout=10, fake_errors=False):
        self.ip_address = ip_address
        self.port = int(port)
        self.fake_errors = fake_errors

        self.model = None
        self.serial = None
        self.software_revision = None

        self.comm = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.comm.connect((self.ip_address, self.port))  # connects to the PTC
        self.comm.settimeout(timeout)

    def get_data(self):
        """
        Gets the raw data from the ptc and returns it in a usable format.
        """
        self.comm.sendall(self.buildRegistersQuery())
        data = self.comm.recv(1024)
        data_flag, brd = self.breakdownReplyData(data)

        return data_flag, brd

    @staticmethod
    def buildRegistersQuery():
        query = bytes([0x09, 0x99,  # Message ID
                       0x00, 0x00,  # Unused
                       0x00, 0x06,  # Message size in bytes
                       0x01,        # Slave Address
                       0x04,        # Function Code
                       0x00, 0x01,   # The starting Register Number
                       0x00, 0x35])  # How many to read
        return query

    def power(self, state):
        """Turn the PTC on or off.

        Parameters
        ----------
        state : str
            Desired power state of the PTC, either 'on', or 'off'.

        """
        command = [0x09, 0x99,  # Message ID
                   0x00, 0x00,  # Unused
                   0x00, 0x06,  # Message size in bytes
                   0x01,        # Slave Address
                   0x06,        # Function Code
                   0x00, 0x01]   # Register Number

        if state.lower() == 'on':
            command.extend([0x00, 0x01])
        elif state.lower() == 'off':
            command.extend([0x00, 0xff])
        else:
            raise ValueError(f"Invalid state: {state}")

        self.comm.sendall(bytes(command))
        self.comm.recv(1024)  # Discard the echoed command

    def breakdownReplyData(self, rawdata):
        """Take in raw ptc data, and return a dictionary.

        The dictionary keys are the data labels, the dictionary values are the
        data in floats or ints.

        Returns
        -------
        data_flag : bool
            False if data is valid, True if output could not be interpretted.
        data : dict
            Data dictionary already formatted for passing to OCS Feed.

        """

        # Associations between keys and their location in rawData
        keyloc = {"Operating_State": [9, 10],
                  "Compressor_State": [11, 12],
                  "Warning_State": [15, 16, 13, 14],
                  "Alarm_State": [19, 20, 17, 18],
                  "Coolant_In_Temp": [23, 24, 21, 22],
                  "Coolant_Out_Temp": [27, 28, 25, 26],
                  "Oil_Temp": [31, 32, 29, 30],
                  "Helium_Temp": [35, 36, 33, 34],
                  "Low_Pressure": [39, 40, 37, 38],
                  "Low_Pressure_Average": [43, 44, 41, 42],
                  "High_Pressure": [47, 48, 45, 46],
                  "High_Pressure_Average": [51, 52, 49, 50],
                  "Delta_Pressure_Average": [55, 56, 53, 54],
                  "Motor_Current": [59, 60, 57, 58],
                  "Hours_of_Operation": [63, 64, 61, 62],
                  "Pressure_Unit": [65, 66],
                  "Temperature_Unit": [67, 68],
                  "Serial_Number": [69, 70],
                  "Model": [71, 72],
                  "Software_Revision": [73, 74]}

        # Iterate through all keys and return the data in a usable format.
        # If there is an error in the string format, print the
        # error to logs, return an empty dictionary, and flag the data as bad
        data = {}

        # If fake_errors=True, then randomly output the string 'FAKE ERROR'
        # instead of the actual data 50% of the time
        if self.fake_errors:
            if random.random() < 0.5:
                rawdata = "FAKE ERROR"

        try:
            for key in keyloc.keys():
                locs = keyloc[key]
                wkrBytes = bytes([rawdata[loc] for loc in locs])

                # four different data formats to unpack
                # Big endian unsigned integer 16 bits
                if key in [
                    "Operating_State",
                    "Compressor_State",
                    "Pressure_Unit",
                    "Temperature_Unit",
                    "Serial_Number",
                ]:
                    state = struct.unpack(">H", wkrBytes)[0]
                    # Serial number is an attribute, not publishable data
                    if key == "Serial_Number":
                        self.serial = state
                    else:
                        data[key] = state
                # 32bit signed integer which is actually stored as a
                # 32bit IEEE float (silly)
                elif key in ["Warning_State", "Alarm_State"]:
                    state = int(struct.unpack(">f", wkrBytes)[0])
                    data[key] = state
                # 2 x 8-bit lookup tables.
                elif key in ["Model"]:
                    model_major = struct.unpack(
                        ">B", bytes([rawdata[locs[0]]]))[0]
                    model_minor = struct.unpack(
                        ">B", bytes([rawdata[locs[1]]]))[0]
                    # Model is an attribute, not publishable data
                    self.model = str(model_major) + "_" + str(model_minor)
                elif key in ["Software_Revision"]:
                    version_major = struct.unpack(
                        ">B", bytes([rawdata[locs[0]]]))[0]
                    version_minor = struct.unpack(
                        ">B", bytes([rawdata[locs[1]]]))[0]
                    self.software_revision = str(version_major) + "." + str(version_minor)
                # 32 bit Big endian IEEE floating point
                else:
                    data[key] = struct.unpack(">f", wkrBytes)[0]

            data_flag = False

        except BaseException:
            data_flag = True
            print("Compressor output could not be converted to numbers."
                  f"Skipping this data block. Bad output string is {rawdata}")

        return data_flag, data

    def __del__(self):
        """
        If the PTC class instance is destroyed, close the connection to the
        ptc.
        """
        self.comm.close()


class PTCAgent:
    """Agent to connect to a single cryomech compressor.

    Parameters:
        port (int): TCP port to connect to.
        ip_address (str): IP Address for the compressor.
        f_sample (float, optional): Data acquisiton rate, defaults to 2.5 Hz.
        fake_errors (bool, optional): Generates fake errors in the string
            output 50% of the time.

    """

    def __init__(self, agent, port, ip_address, f_sample=2.5,
                 fake_errors=False):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.ip_address = ip_address
        self.fake_errors = fake_errors

        self.port = port
        self.module = None
        self.f_sample = f_sample

        self.initialized = False
        self.take_data = False

        # Registers data feeds
        agg_params = {
            'frame_length': 60,
        }
        self.agent.register_feed('ptc_status',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params=None):
        """init(auto_acquire=False)

        **Task** - Initializes the connection to the PTC.

        Parameters:
            auto_acquire (bool): Automatically start acq process after
                initialization if True. Defaults to False.

        """
        if self.initialized:
            return True, "Already Initialized"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            # Establish connection to ptc
            self.ptc = PTC(self.ip_address, port=self.port,
                           fake_errors=self.fake_errors)

            # Test connection and display identifying info
            self.ptc.get_data()
            print("PTC Model:", self.ptc.model)
            print("PTC Serial Number:", self.ptc.serial)
            print("Software Revision is:", self.ptc.software_revision)

        self.initialized = True

        # Start data acquisition if requested
        if params['auto_acquire']:
            resp = self.agent.start('acq', params={})
            self.log.info(f'Response from acq.start(): {resp[1]}')

        return True, "PTC agent initialized"

    @ocs_agent.param('state', type=str, choices=['off', 'on'])
    def power_ptc(self, session, params=None):
        """power_ptc(state=None)

        **Task** - Remotely turn the PTC on or off.

        Parameters
        ----------
        state : str
            Desired power state of the PTC, either 'on', or 'off'.

        """
        with self.lock.acquire_timeout(3, job='power_ptc') as acquired:
            if not acquired:
                self.log.warn("Could not start task because {} is already "
                              "running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            self.ptc.power(params['state'])

        return True, "PTC powered {}".format(params['state'])

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params):
        """acq()

        **Process** - Starts acqusition of data from the PTC.

        Parameters:
            test_mode (bool, optional): Run the Process loop only once.
                This is meant only for testing. Default is False.

        """
        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already"
                              "running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            last_release = time.time()

            self.take_data = True

            while self.take_data:
                # Relinquish sampling lock occasionally
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self.lock.job}.")
                        continue

                # Publish data, waiting 1/f_sample seconds in between calls.
                pub_data = {'timestamp': time.time(),
                            'block_name': 'ptc_status'}
                data_flag, data = self.ptc.get_data()
                pub_data['data'] = data
                # If there is an error in compressor output (data_flag = True),
                # do not publish
                if not data_flag:
                    self.agent.publish_to_feed('ptc_status', pub_data)
                time.sleep(1. / self.f_sample)

                if params['test_mode']:
                    break

            self.agent.feeds["ptc_status"].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params):
        """Stops acqusition of data from the PTC."""
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--port', default=502)
    pgroup.add_argument('--serial-number')
    pgroup.add_argument('--mode', choices=['init', 'acq'])
    pgroup.add_argument('--fake-errors', default=False,
                        help="If True, randomly output 'FAKE ERROR' instead of "
                             "data half of the time.")

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='CryomechCPAAgent',
                                  parser=parser,
                                  args=args)
    print('I am in charge of device with serial number: %s' % args.serial_number)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    # Call launcher function (initiates connection to appropriate
    # WAMP hub and realm).

    agent, runner = ocs_agent.init_site_agent(args)

    # create agent instance and run log creation
    ptc = PTCAgent(agent, args.port, args.ip_address,
                   fake_errors=args.fake_errors)

    agent.register_task('init', ptc.init, startup=init_params)
    agent.register_process('acq', ptc.acq, ptc._stop_acq)
    agent.register_task('power_ptc', ptc.power_ptc)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



"""OCS agent module to read the data from beagleboneblack for CHWP encoder

Note
----
   This is confirmed to be work with the following versions of beagleboneblack software:
   - currently hwpdaq branch in spt3g_software_sa repository: 731ff39
   (sha256sum)
   - Encoder1.bin: c8281525bdd0efae66aede7cffc3520ab719cfd67f6c2d7fd01509a4289a9d32
   - Encoder2.bin: a6ed9d89e9cf26036bf1da9e7e2098da85bbfa6eb08d0caef1e3c40877dd5077
   - IRIG1.bin: 7bc37b30a1759eb792f0db176bcd6080f9c3c7ec78ba2e1614166b2031416091
   - IRIG2.bin: d206dd075f73c32684d8319c9ed19f019cc705a9f253725de085eb511b8c0a12

Data feeds
----------
HWPEncoder:
   (HWPEncoder_counter_sub)
   counter_sub: subsampled counter values [::NUM_SUBSAMPLE]
   counter_index_sub: subsampled index counter values

   (HWPEncoder_freq)
   approx_hwp_freq: approximate estimate of hwp rotation frequency
   diff_counter_mean: mean of diff(counter)
   diff_index_mean: mean of diff(counter_index)
   diff_counter_std: std of diff(counter)
   diff_index_std: std of diff(counter_index)

   (HWPEncoder_quad)
   quad: quadrature data

   (HWPEncoder_irig)
   irig_time: decoded time in second since the unix epoch
   rising_edge_cont: BBB clcok count values
                     for the IRIG on-time reference marker risinge edge
   irig_sec: seconds decoded from IRIG-B
   irig_min: minutes decoded from IRIG-B
   irig_hour: hours decoded from IRIG-B
   irig_day: days decoded from IRIG-B
   irig_year: years decoded from IRIG-B
   bbb_clock_freq: BBB clock frequency estimate using IRIG-B

   (HWPEncoder_irig_raw)
   irig_synch_pulse_clock_time: reference marker time in sec
   irig_synch_pulse_clock_counts: clock counts for reference markers
   irig_info: IRIG bit info

HWPEncoder_full: separated feed for full-sample HWP encoder data,
                 not to be included in influxdb database
   (HWPEncoder_counter)
   counter: BBB counter values for encoder signal edges
   counter_index: index numbers for detected edges by BBB
"""

import argparse
import calendar
import select
import socket
import struct
import time
from collections import deque

import numpy as np
import txaio

txaio.use_twisted()

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

# These three values (COUNTER_INFO_LENGTH, COUNTER_PACKET_SIZE, IRIG_PACKET_SIZE)
# should be consistent with the software on beaglebone.
# The number of datapoints in every encoder packet from the Beaglebone
COUNTER_INFO_LENGTH = 120
# The size of the encoder packet from the beaglebone
#    (header + 3*COUNTER_INFO_LENGTH datapoint information + 1 quadrature readout)
COUNTER_PACKET_SIZE = 4 + 4 * COUNTER_INFO_LENGTH + 8 * COUNTER_INFO_LENGTH + 4
# The size of the IRIG packet from the Beaglebone
IRIG_PACKET_SIZE = 132

# The slit scaler value for rough HWP rotating frequency
NUM_SLITS = 570
# Number of encoder counter samples to publish at once
NUM_ENCODER_TO_PUBLISH = 4200
# Seconds to publish encoder data even before reaching NUM_ENCODER_TO_PUBLISH
SEC_ENCODER_TO_PUBLISH = 10
# Subsampling facot for the encoder counter data to influxdb
NUM_SUBSAMPLE = 500


def de_irig(val, base_shift=0):
    """Converts the IRIG signal into sec/min/hours/day/year depending on the parameters

    Parameters
    ----------
    val : int
       raw IRIG bit info of each 100msec chunk
    base_shift : int, optional
       number of bit shifts. This should be 0 except for seccods

    Returns
    -------
    int
       Either of sec/min/hourds/day/year

    """
    return (((val >> (0 + base_shift)) & 1)
            + ((val >> (1 + base_shift)) & 1) * 2
            + ((val >> (2 + base_shift)) & 1) * 4
            + ((val >> (3 + base_shift)) & 1) * 8
            + ((val >> (5 + base_shift)) & 1) * 10
            + ((val >> (6 + base_shift)) & 1) * 20
            + ((val >> (7 + base_shift)) & 1) * 40
            + ((val >> (8 + base_shift)) & 1) * 80)


def count2time(counts, t_offset=0.):
    """Quick etimation of time using Beagleboneblack clock counts

    Parameters
    ----------
    counts : list of int
       Beagleboneblack clock counter value
    t_offset : int, optional
       time offset in seconds

    Returns
    -------
    list of float
       Estimated time in seconds assuming the Beagleboneblack clock frequency is 200 MHz.
       Without specifying t_offset, output is just the difference
       from the first sample in the input list

    """
    t_array = np.array(counts, dtype=float) - counts[0]
    # Assuming BBB clock is 200MHz
    t_array *= 5.e-9
    t_array += t_offset

    return t_array.tolist()


class EncoderParser:
    """Class which will parse the incoming packets from the BeagleboneBlack and store the data

    Attributes
    ----------
    counter_queue : deque object
       deque to store the encoder counter data
    irig_queue : deque object
       deque to store the IRIG data
    is_start : int
       Used for procedures that only run when data collection begins
       Initialized to be 1, until the first IRIG parsing happens and set to 0
    start_time : list of int
       Will hold the time at which data collection started [hours, mins, secs]
    current_time : int
       Current unix timestamp in seconds parased from IRIG
    sock : scoket.sock
       a UDP socket to connect to the Beagleboneblack
    data : str
       String which will hold the raw data from the Beaglebone before it is parsed
    read_chunk_size : int
       Maximum data size to receive UDP packets in bytes

   Parameters
    ----------
    beaglebone_port : int, optional
       Port number to receive UDP packets from Beagleboneblack
       This must be the same as the localPort in the Beaglebone code
    read_chunk_size : int, optional
       Maximum data size to receive UDP packets in bytes
       read_chunk_size: This value shouldn't need to change

    """

    def __init__(self, beaglebone_port=8080, read_chunk_size=8196):
        # Creates twoe queues to hold the data from the encoder, IRIG, and quadrature respectively
        self.counter_queue = deque()
        self.irig_queue = deque()

        # Used for procedures that only run when data collection begins
        self.is_start = 1
        # Will hold the time at which data collection started [hours, mins, secs]
        self.start_time = [0, 0, 0]
        # Will be continually updated with unix in seconds
        self.current_time = 0

        # Creates a UDP socket to connect to the Beaglebone
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # Binds the socket to a specific ip address and port
        # The ip address can be blank for accepting any UDP packet to the port
        self.sock.bind(('', beaglebone_port))
        # self.sock.setblocking(0)

        # String which will hold the raw data from the Beaglebone before it is parsed
        self.data = ''
        self.read_chunk_size = read_chunk_size

        self.log = txaio.make_logger()

    def pretty_print_irig_info(self, irig_info, edge, print_out=False):
        """Takes the IRIG information, prints it to the screen, sets the current time,
        and returns the current time

        Parameters
        ----------
        irig_info : list of int
           IRIG bit info
        edge : int
           Clock count of rising edge of a reference marker bit
        print_out : bool, optional
           Set True to print out the parsed timestamp

        Returns
        -------
        current_time : int
           Current unix timestamp in seconds parased from IRIG

        """
        # Calls self.de_irig() to get the sec/min/hour of the IRIG packet
        secs = de_irig(irig_info[0], 1)
        mins = de_irig(irig_info[1], 0)
        hours = de_irig(irig_info[2], 0)
        day = de_irig(irig_info[3], 0) \
            + de_irig(irig_info[4], 0) * 100
        year = de_irig(irig_info[5], 0)

        # If it is the first time that the function is called then set self.start_time
        # to the current time
        if self.is_start == 1:
            self.start_time = [hours, mins, secs]
            self.is_start = 0

        if print_out:
            # Find the sec/min/hour digit difference from the start time
            dsecs = secs - self.start_time[2]
            dmins = mins - self.start_time[1]
            dhours = hours - self.start_time[0]

            # Corrections to make sure that dsecs/dmins/dhours are all positive
            if dhours < 0:
                dhours = dhours + 24

            if (dmins < 0) or ((dmins == 0) and (dsecs < 0)):
                dmins = dmins + 60
                dhours = dhours - 1

            if dsecs < 0:
                dsecs = dsecs + 60
                dmins = dmins - 1

            # Print UTC time, run time, and current clock count of the beaglebone
            print('Current Time:', ('%d:%d:%d' % (hours, mins, secs)),
                  'Run Time', ('%d:%d:%d' % (dhours, dmins, dsecs)),
                  'Clock Count', edge)

        # Set the current time in seconds (changed to seconds from unix epoch)
        # self.current_time = secs + mins*60 + hours*3600
        try:
            st_time = time.strptime("%d %d %d:%d:%d" % (year, day, hours, mins, secs),
                                    "%y %j %H:%M:%S")
            self.current_time = calendar.timegm(st_time)
        except ValueError:
            self.log.error(f'Invalid IRIG-B timestamp: {year} {day} {hours} {mins} {secs}')
            self.current_time = -1

        return self.current_time

    def check_data_length(self, start_index, size_of_read):
        """Checks to make sure that self.data is the right size
        Return false if the wrong size, return true if the data is the right size

        Parameters
        ----------
        start_index : int
           first index of the data to read
        size_of_read : int
           data size to read in bytes

        Returns
        -------
        bool
           False if the current data size is smaller than the data size suggested by the header info

        """
        if start_index + size_of_read > len(self.data):
            self.data = self.data[start_index:]
            return False

        return True

    def grab_and_parse_data(self):
        """Grabs self.data, determine what packet it corresponds to, parses the data.
        This is a while loop to look for an appropriate header in a packet from beaglebone.
        Then, the data will be passed to an appropriate parsing method
        and stored in either of counter_queue or irig_queue.
        The detailed structure of the queues can be found in parse_counter_info/parse_irig_info.

        If unexpected data length found, this will output some messages:
           Error 0: data length is shorter than the header size (4 bytes)
           Error 1: data length is shorter than the encoder counter info
                    even though the encoder packet header is found.
           Error 2: data length is shorter than the IRIG info
                    even though the IRIG packet header is found.
        """
        while True:
            # If there is data from the socket attached to the beaglebone then
            #     ready[0] = true
            # If not then continue checking for 2 seconds and if there is still no data
            #     ready[0] = false
            ready = select.select([self.sock], [], [], 2)
            if ready[0]:
                # Add the data from the socket attached to the beaglebone
                # to the self.data string
                data = self.sock.recv(self.read_chunk_size)
                if len(self.data) > 0:
                    self.data += data
                else:
                    self.data = data

                while True:
                    # Check to make sure that there is at least 1 int in the packet
                    # The first int in every packet should be the header
                    if not self.check_data_length(0, 4):
                        self.log.error('Error 0')
                        break

                    header = self.data[0:4]
                    # Convert a structure value from the beaglebone (header) to an int
                    header = struct.unpack('<I', header)[0]
                    # print('header ', '0x%x'%header)

                    # 0x1EAF = Encoder Packet
                    # 0xCAFE = IRIG Packet
                    # 0xE12A = Error Packet

                    # Encoder
                    if header == 0x1eaf:
                        # Make sure the data is the correct length for an Encoder Packet
                        if not self.check_data_length(0, COUNTER_PACKET_SIZE):
                            self.log.error('Error 1')
                            break
                        # Call the meathod self.parse_counter_info() to parse the Encoder Packet
                        self.parse_counter_info(self.data[4: COUNTER_PACKET_SIZE])
                        if len(self.data) >= COUNTER_PACKET_SIZE:
                            self.data = self.data[COUNTER_PACKET_SIZE:]

                    # IRIG
                    elif header == 0xcafe:
                        # Make sure the data is the correct length for an IRIG Packet
                        if not self.check_data_length(0, IRIG_PACKET_SIZE):
                            self.log.error('Error 2')
                            break
                        # Call the meathod self.parse_irig_info() to parse the IRIG Packet
                        self.parse_irig_info(self.data[4: IRIG_PACKET_SIZE])
                        if len(self.data) >= IRIG_PACKET_SIZE:
                            self.data = self.data[IRIG_PACKET_SIZE:]

                    # Error
                    # An Error Packet will be sent if there is a timing error in the
                    # synchronization pulses of the IRIG packet
                    # If you see 'Packet Error' check to make sure the IRIG is functioning as
                    # intended and that all the connections are made correctly
                    elif header == 0xe12a:
                        self.log.error('Packet Error')
                        # Clear self.data
                        self.data = ''
                    elif header == 0x1234:
                        self.log.error('Received timeout packet.')
                        # Clear self.data
                        self.data = ''
                    else:
                        self.log.error('Bad header')
                        # Clear self.data
                        self.data = ''

                    if len(self.data) == 0:
                        break
                break

            # If there is no data from the beaglebone 'Looking for data ...' will print
            # If you see this make sure that the beaglebone has been set up properly
            # print('Looking for data ...')

    def parse_counter_info(self, data):
        """Method to parse the Encoder Packet and put them to counter_queue

        Parameters
        ----------
        data : str
           string for the encoder ounter info

        Note:
           'data' structure:
           (Please note that '150' below might be replaced by COUNTER_INFO_LENGTH)
           [0] Readout from the quadrature
           [1-150] clock counts of 150 data points
           [151-300] corresponding clock overflow of the 150 data points (each overflow count
           is equal to 2^16 clock counts)
           [301-450] corresponding absolute number of the 150 data points ((1, 2, 3, etc ...)
           or (150, 151, 152, etc ...) or (301, 302, 303, etc ...) etc ...)

           counter_queue structure:
           counter_queue = [[64 bit clock counts],
                            [clock count indicese incremented by every edge],
                            quadrature,
                            current system time]
        """

        # Convert the Encoder Packet structure into a numpy array
        derter = np.array(struct.unpack('<' + 'I' + 'III' * COUNTER_INFO_LENGTH, data))

        # self.quad_queue.append(derter[0].item()) # merged to counter_queue
        self.counter_queue.append((derter[1:COUNTER_INFO_LENGTH + 1]
                                   + (derter[COUNTER_INFO_LENGTH + 1:2 * COUNTER_INFO_LENGTH + 1] << 32),
                                   derter[2 * COUNTER_INFO_LENGTH + 1:3 * COUNTER_INFO_LENGTH + 1],
                                   derter[0].item(), time.time()))

    def parse_irig_info(self, data):
        """Method to parse the IRIG Packet and put them to the irig_queue

        Parameters
        ----------
        data : str
           string for the IRIG info

        Note
        ----
           'data' structure:
           [0] clock count of the IRIG Packet which the UTC time corresponds to
           [1] overflow count of initial rising edge
           [2] binary encoding of the second data
           [3] binary encoding of the minute data
           [4] binary encoding of the hour data
           [5-11] additional IRIG information which we do mot use
           [12-21] synchronization pulse clock counts
           [22-31] overflow count at each synchronization pulse

           irig_queue structure:
           irig_queue = [Packet clock count,
                         Packet UTC time in sec,
                         [binary encoded IRIG data],
                         [synch pulses clock counts],
                         current system time]

        """

        # Convert the IRIG Packet structure into a numpy array
        unpacked_data = struct.unpack('<L' + 'L' + 'L' * 10 + 'L' * 10 + 'L' * 10, data)

        # Start of the packet clock count
        # overflow.append(unpacked_data[1])
        # print "overflow: ", overflow

        rising_edge_time = unpacked_data[0] + (unpacked_data[1] << 32)

        # Stores IRIG time data
        irig_info = unpacked_data[2:12]

        # Prints the time information and returns the current time in seconds
        irig_time = self.pretty_print_irig_info(irig_info, rising_edge_time)

        # Stores synch pulse clock counts accounting for overflow of 32 bit counter
        synch_pulse_clock_times = (np.asarray(unpacked_data[12:22])
                                   + (np.asarray(unpacked_data[22:32]) << 32)).tolist()

        # self.irig_queue = [Packet clock count,Packet UTC time in sec,
        #                    [binary encoded IRIG data],[synch pulses clock counts],
        #                    [current system time]]
        self.irig_queue.append((rising_edge_time, irig_time, irig_info,
                                synch_pulse_clock_times, time.time()))

    def __del__(self):
        self.sock.close()


class HWPBBBAgent:
    """OCS agent for HWP encoder DAQ using Beaglebone Black

    Attributes
    ----------
    rising_edge_count : int
       clock count values for the rising edge of IRIG reference marker,
       saved for calculating the beaglebone clock frequency
    irig_time : int
       unix timestamp from IRIG

    """

    def __init__(self, agent_obj, port=8080):
        self.active = True
        self.agent = agent_obj
        self.log = agent_obj.log
        self.lock = TimeoutLock()
        self.port = port
        self.take_data = False
        self.initialized = False
        # For clock count to time conversion
        self.rising_edge_count = 0
        self.irig_time = 0

        agg_params = {'frame_length': 60}
        self.agent.register_feed('HWPEncoder', record=True,
                                 agg_params=agg_params)
        agg_params = {'frame_length': 60, 'exclude_influx': True}
        self.agent.register_feed('HWPEncoder_full', record=True,
                                 agg_params=agg_params)
        self.parser = EncoderParser(beaglebone_port=self.port)

    def acq(self, session, params):
        """acq()

        **Process** - Start acquiring data.

        """
        time_encoder_published = 0
        counter_list = []
        counter_index_list = []
        quad_list = []
        quad_counter_list = []
        received_time_list = []

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn('Could not start acq because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock.'

            session.set_status('running')

            self.take_data = True

            # Prepare data_cache for session.data
            self.hwp_freq = -1
            self.ct = time.time()
            data_cache = {
                'approx_hwp_freq': self.hwp_freq,
                'encoder_last_updated': self.ct,
                'irig_time': self.irig_time,
                'irig_last_updated': self.ct,
            }

            while self.take_data:
                # This is blocking until data are available
                self.parser.grab_and_parse_data()

                # IRIG data; normally every sec
                while len(self.parser.irig_queue):
                    irig_data = self.parser.irig_queue.popleft()
                    rising_edge_count = irig_data[0]
                    irig_time = irig_data[1]
                    irig_info = irig_data[2]
                    synch_pulse_clock_counts = irig_data[3]
                    sys_time = irig_data[4]
                    data = {'timestamp': sys_time, 'block_name': 'HWPEncoder_irig', 'data': {}}
                    data['data']['irig_time'] = irig_time
                    data['data']['rising_edge_count'] = rising_edge_count
                    data['data']['irig_sec'] = de_irig(irig_info[0], 1)
                    data['data']['irig_min'] = de_irig(irig_info[1], 0)
                    data['data']['irig_hour'] = de_irig(irig_info[2], 0)
                    data['data']['irig_day'] = de_irig(irig_info[3], 0) \
                        + de_irig(irig_info[4], 0) * 100
                    data['data']['irig_year'] = de_irig(irig_info[5], 0)

                    # Beagleboneblack clock frequency measured by IRIG
                    if self.rising_edge_count > 0 and irig_time > 0:
                        bbb_clock_freq = float(rising_edge_count - self.rising_edge_count) \
                            / (irig_time - self.irig_time)
                    else:
                        bbb_clock_freq = 0.
                    data['data']['bbb_clock_freq'] = bbb_clock_freq

                    self.agent.publish_to_feed('HWPEncoder', data)
                    self.rising_edge_count = rising_edge_count
                    self.irig_time = irig_time

                    # saving clock counts for every refernce edge and every irig bit info
                    data = {'timestamps': [], 'block_name': 'HWPEncoder_irig_raw', 'data': {}}
                    # 0.09: time difference in seconds b/w reference marker and
                    #       the first index marker
                    data['timestamps'] = sys_time + 0.09 + np.arange(10) * 0.1
                    data['data']['irig_synch_pulse_clock_time'] = list(irig_time + 0.09
                                                                       + np.arange(10) * 0.1)
                    data['data']['irig_synch_pulse_clock_counts'] = synch_pulse_clock_counts
                    data['data']['irig_info'] = list(irig_info)
                    self.agent.publish_to_feed('HWPEncoder', data)

                    data_cache['irig_time'] = self.irig_time
                    data_cache['irig_last_updated'] = sys_time
                    session.data.update(data_cache)

                # Reducing the packet size, less frequent publishing
                # Encoder data; packet coming rate = 570*2*2/150/4 ~ 4Hz packet at 2 Hz rotation
                while len(self.parser.counter_queue):
                    counter_data = self.parser.counter_queue.popleft()

                    counter_list += counter_data[0].tolist()
                    counter_index_list += counter_data[1].tolist()

                    quad_data = counter_data[2]
                    sys_time = counter_data[3]

                    received_time_list.append(sys_time)
                    quad_list.append(quad_data)
                    quad_counter_list.append(counter_data[0][0])
                    ct = time.time()

                    if len(counter_list) >= NUM_ENCODER_TO_PUBLISH \
                       or (len(counter_list)
                           and (ct - time_encoder_published) > SEC_ENCODER_TO_PUBLISH):
                        # Publishing quadratic data first
                        data = {'timestamps': [], 'block_name': 'HWPEncoder_quad', 'data': {}}
                        data['timestamps'] = received_time_list
                        data['data']['quad'] = quad_list
                        self.agent.publish_to_feed('HWPEncoder', data)

                        # Publishing counter data
                        # (full sampled data will not be recorded in influxdb)
                        data = {'timestamps': [], 'block_name': 'HWPEncoder_counter', 'data': {}}
                        data['data']['counter'] = counter_list
                        data['data']['counter_index'] = counter_index_list

                        data['timestamps'] = count2time(counter_list, received_time_list[0])
                        self.agent.publish_to_feed('HWPEncoder_full', data)

                        # Subsampled data for influxdb display
                        data_subsampled = {'block_name': 'HWPEncoder_counter_sub', 'data': {}}
                        data_subsampled['timestamps'] = np.array(data['timestamps'])[::NUM_SUBSAMPLE].tolist()
                        data_subsampled['data']['counter_sub'] = np.array(counter_list)[::NUM_SUBSAMPLE].tolist()
                        data_subsampled['data']['counter_index_sub'] = np.array(counter_index_list)[::NUM_SUBSAMPLE].tolist()
                        self.agent.publish_to_feed('HWPEncoder', data_subsampled)

                        # For rough estimation of HWP rotation frequency
                        data = {'timestamp': received_time_list[0],
                                'block_name': 'HWPEncoder_freq', 'data': {}}
                        dclock_counter = counter_list[-1] - counter_list[0]
                        dindex_counter = counter_index_list[-1] - counter_index_list[0]
                        # Assuming Beagleboneblack clock is 200 MHz
                        pulse_rate = dindex_counter * 2.e8 / dclock_counter
                        hwp_freq = pulse_rate / 2. / NUM_SLITS

                        diff_counter = np.diff(counter_list)
                        diff_index = np.diff(counter_index_list)

                        self.log.info(f'pulse_rate {pulse_rate} {hwp_freq}')
                        data['data']['approx_hwp_freq'] = hwp_freq
                        data['data']['diff_counter_mean'] = np.mean(diff_counter)
                        data['data']['diff_index_mean'] = np.mean(diff_index)
                        data['data']['diff_counter_std'] = np.std(diff_counter)
                        data['data']['diff_index_std'] = np.std(diff_index)
                        self.agent.publish_to_feed('HWPEncoder', data)

                        # Initialize lists
                        counter_list = []
                        counter_index_list = []
                        quad_list = []
                        quad_counter_list = []
                        received_time_list = []

                        time_encoder_published = ct

                        # Update session.data
                        self.hwp_freq = hwp_freq
                        self.ct = ct

                data_cache['approx_hwp_freq'] = self.hwp_freq
                data_cache['encoder_last_updated'] = self.ct
                session.data.update(data_cache)

        self.agent.feeds['HWPEncoder'].flush_buffer()
        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops the data acquisiton.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'

        return False, 'acq is not currently running.'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', type=int, default=8080)

    return parser


# Portion of the code that runs
def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPBBBAgent',
                                  parser=parser,
                                  args=args)
    agent, runner = ocs_agent.init_site_agent(args)
    hwp_bbb_agent = HWPBBBAgent(agent, port=args.port)
    agent.register_process('acq', hwp_bbb_agent.acq, hwp_bbb_agent._stop_acq, startup=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import socket
import time
from os import environ

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.pfeiffer_tc400.drivers import PfeifferTC400


class PfeifferTC400Agent:
    """Agent to connect to a pfeiffer tc400 electronic drive unit controlling a
    turbo pump via a serial-to-ethernet converter.

    Parameters
    ----------
    ip_address : str
        IP address for the serial-to-ethernet converter
    port_number : int
        Serial-to-ethernet converter port
    turbo_address : int
        An internal address used to communicate between the power supplies and
        the tc400. Found on the front screen of the power supplies.
    """

    def __init__(self, agent, ip_address, port_number, turbo_address):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.job = None
        self.ip_address = ip_address
        self.port_number = port_number
        self.turbo_address = turbo_address
        self.monitor = False

        self.turbo = None

        # Registers Temperature and Voltage feeds
        agg_params = {
            'frame_length': 10 * 60,
        }
        self.agent.register_feed('pfeiffer_turbo',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params):
        """init(auto_acquire=False)

        **Task** - Initialize the connection to the turbo controller.

        Parameters
        ----------
        auto_acquire: bool, optional
            Default is False. Starts data acquisition after initialization
            if True.
        """

        with self.lock.acquire_timeout(0) as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            try:
                self.turbo = PfeifferTC400(self.ip_address,
                                           self.port_number,
                                           self.turbo_address)
            except socket.timeout as e:
                self.log.error("Turbo Controller timed out"
                               + f"during connect with error {e}")
                return False, "Timeout"
            self.log.info("Connected to turbo controller")

        # Start data acquisition if requested in site-config
        auto_acquire = params.get('auto_acquire', False)
        if auto_acquire:
            self.agent.start('acq')

        return True, 'Initialized Turbo Controller.'

    @ocs_agent.param('test_mode', default=False, type=bool)
    @ocs_agent.param('wait', default=1, type=float)
    def acq(self, session, params):
        """acq(wait=1, test_mode=False)

        **Process** - Continuously monitor turbo motor temp and rotation speed
        and send info to aggregator.

        The ``session.data`` object stores the most recent published values
        in a dictionary. For example::

            session.data = {
                'timestamp': 1598626144.5365012,
                'block_name': 'turbo_output',
                'data': {
                    "Turbo_Motor_Temp": 40.054,
                    "Rotation_Speed": 823.655,
                    "Error_Code": "Err001",
                }
            }

        Parameters
        ----------
        wait: float, optional
            time to wait between measurements [seconds]. Default=1s.

        """
        self.monitor = True

        while self.monitor:
            with self.lock.acquire_timeout(1) as acquired:
                if acquired:
                    data = {
                        'timestamp': time.time(),
                        'block_name': 'turbo_output',
                        'data': {}
                    }

                    try:
                        data['data']["Turbo_Motor_Temp"] = self.turbo.get_turbo_motor_temperature()
                        data['data']["Rotation_Speed"] = self.turbo.get_turbo_actual_rotation_speed()
                        data['data']['Error_Code'] = self.turbo.get_turbo_error_code()
                    except ValueError as e:
                        self.log.error(f"Error in collecting data: {e}")
                        continue

                    self.agent.publish_to_feed('pfeiffer_turbo', data)

                    # Allow this process to be queried to return current data
                    session.data = data

                else:
                    self.log.warn("Could not acquire in monitor turbo")

            time.sleep(params['wait'])

            if params['test_mode']:
                break

        return True, "Finished monitoring turbo"

    def _stop_acq(self, session, params):
        """Stop monitoring the turbo output."""
        if self.monitor:
            self.monitor = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('_')
    def turn_turbo_on(self, session, params):
        """turn_turbo_on()

        **Task** - Turns the turbo on.

        """

        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                ready = self.turbo.ready_turbo()
                if not ready:
                    return False, "Setting to ready state failed"
                time.sleep(1)
                on = self.turbo.turn_turbo_motor_on()
                if not on:
                    return False, "Turbo unable to be turned on"
            else:
                return False, "Could not acquire lock"

        return True, 'Turned turbo on'

    @ocs_agent.param('_')
    def turn_turbo_off(self, session, params):
        """turn_turbo_off()

        **Task** - Turns the turbo off.

        """

        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                off = self.turbo.turn_turbo_motor_off()
                if not off:
                    return False, "Turbo unable to be turned off"
                time.sleep(1)
                unready = self.turbo.unready_turbo()
                if not unready:
                    return False, "Setting to ready state failed"
            else:
                return False, "Could not acquire lock"

        return True, 'Turned turbo off'

    @ocs_agent.param('_')
    def acknowledge_turbo_errors(self, session, params):
        """acknowledge_turbo_errors()

        **Task** - Sends an acknowledgment of the error code to the turbo.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.turbo.acknowledge_turbo_errors()
            else:
                return False, "Could not acquire lock"

        return True, 'Acknowledged Turbo Errors.'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address', type=str, help="Serial-to-ethernet "
                        + "converter ip address")
    pgroup.add_argument('--port-number', type=int, help="Serial-to-ethernet "
                        + "converter port")
    pgroup.add_argument('--turbo-address', type=int, help="Internal address "
                        + "used by power supplies")
    pgroup.add_argument('--mode', type=str, help="Set to acq to run acq on "
                        + "startup")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = site_config.add_arguments()

    # Get the default ocs agrument parser
    parser = make_parser()
    args = site_config.parse_args(agent_class='PfeifferTC400Agent',
                                  parser=parser,
                                  args=args)

    init_params = False
    if args.mode == 'acq':
        init_params = {'auto_acquire': True}

    agent, runner = ocs_agent.init_site_agent(args)

    p = PfeifferTC400Agent(agent,
                           args.ip_address,
                           int(args.port_number),
                           int(args.turbo_address))

    agent.register_task('init', p.init, startup=init_params)
    agent.register_task('turn_turbo_on', p.turn_turbo_on)
    agent.register_task('turn_turbo_off', p.turn_turbo_off)
    agent.register_task('acknowledge_turbo_errors', p.acknowledge_turbo_errors)
    agent.register_process('acq', p.acq, p._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import json
import os
import queue
import time

from ocs import ocs_agent, ocs_feed, site_config
from twisted.internet import reactor
from twisted.internet.protocol import DatagramProtocol

from socs.db.suprsync import SupRsyncFilesManager, create_file


def create_remote_path(meta, archive_name):
    """
    Creates "remote path" for file.

    For pysmurf ancilliary files (in the ``smurf`` archive), paths are
    generated from the pysmurf action / action timestamp::

        <archive_dir>/<5 ctime digits>/<pub_id>/<action_timestamp>_<action>/<plots or outputs>

    For timestream files, the path will be the relative path of the file
    from  whatever the g3_dir is.

    Args
    -----
        meta (dict):
            A dict containing file metadata that's sent from the pysmurf
            publisher when a new file is registered. Contains info such as the
            pysmurf action, file timestamp, path, publisher id, etc.
        archive_name (str):
            Name of the archive the file belongs to. 'smurf' if it is a pysmurf
            ancilliary file, in which case the path will be generated based on
            the pysmurf action / timestamp

    """
    if archive_name == 'smurf':
        ts = meta['timestamp']
        action_timestamp = meta['action_ts']
        action = meta['action']
        basename = os.path.basename(meta['path'])
        dir_type = 'plots' if meta['plot'] else 'outputs'
        pub_id = meta['pub_id']

        return os.path.join(
            f"{str(ts):.5}",                 # 5 ctime digits
            pub_id,                          # Pysmurf publisher id
            f"{action_timestamp}_{action}",  # grouptime_action
            dir_type,                        # plots/outputs
            basename,
        )
    elif archive_name == 'timestreams':
        return str(os.path.join(*os.path.normpath(meta['path']).split(os.sep)[-3:]))


class PysmurfMonitor(DatagramProtocol):
    """
    Monitor for pysmurf UDP publisher.

    This agent should be run on the smurf-server and will monitor messages
    published via the pysmurf Publisher.

    Main functions are making sure registered files make their way to the pysmurf
    files database, and passing session info to pysmurf-controller agents via
    the ``pysmurf_session_data``

    Args:
        agent (ocs.ocs_agent.OCSAgent):
            OCSAgent object
        args (Namespace):
            argparse namespace with site_config and agent specific arguments

    Attributes:
        agent (ocs.ocs_agent.OCSAgent):
            OCSAgent object
        log (txaio.tx.Logger):
            txaio logger object created by agent
        file_queue (queue.Queue):
            Queue containing metadata for registered files
        db_path (str):
            Path to the suprsync database where files should be entered
        running (bool):
            True if the main process is running.
        echo_sql (bool):
            If True, will echo all sql statements whenever writing to the
            suprsync db.
    """

    def __init__(self, agent, args):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.file_queue = queue.Queue()
        self.db_path = args.db_path
        self.running = False
        self.echo_sql = args.echo_sql

        self.agent.register_feed('pysmurf_session_data')

    def datagramReceived(self, _data, addr):
        """
        Called whenever UDP data is received.

        Args:
            _data (str):
                Raw data passed over UDP port. Pysmurf publisher will send a
                JSON string
            addr (tuple):
                (host, port) of the sender.
        """
        data = json.loads(_data)
        pub_id = data['id']

        if data['type'] in ['data_file', 'g3_file']:
            self.log.info("New file: {fname}", fname=data['payload']['path'])
            data['payload']['pub_id'] = pub_id
            self.file_queue.put(data['payload'])

        elif data['type'] == "session_data" or data['type'] == "session_log":
            self.agent.publish_to_feed(
                "pysmurf_session_data", data, from_reactor=True
            )

        # Handles published metadata from the streamer
        elif data['type'] == "metadata":
            self.log.debug("Received Metadata: {payload}", payload=data['payload'])

            # streamer publisher-id looks like `STREAMER:<stream-id>`
            if ':' in pub_id:
                stream_id = pub_id.split(':')[1]
            else:
                # This is so that this still works before people update to the
                # version of the stream fuction where the pub-id is set
                # properly. In this case the stream-id will be something like
                # "unidentified"
                stream_id = pub_id

            path = data['payload']['path']
            val = data['payload']['value']
            data['payload']['type']

            field_name = ocs_feed.Feed.enforce_field_name_rules(path)
            feed_name = f'{stream_id}_meta'

            if feed_name not in self.agent.feeds:
                self.agent.register_feed(feed_name, record=True, buffer_time=0)

            feed_data = {'block_name': field_name,
                         'timestamp': data['time'],
                         'data': {field_name: val}}

            self.agent.publish_to_feed(feed_name, feed_data, from_reactor=True)

    def run(self, session, params=None):
        """run()

        **Process** - Main process for the pysmurf monitor agent. Processes
        files that have been added to the queue, adding them to the suprsync
        database.
        """
        srfm = SupRsyncFilesManager(self.db_path, create_all=True, echo=self.echo_sql)

        self.running = True
        session.set_status('running')
        while self.running:
            files = []
            while not self.file_queue.empty():
                meta = self.file_queue.get()
                # Archive name defaults to pysmurf because that is currently
                # the only archive. The smurf-streamer will set the
                # archive_name to "timestreams"
                archive_name = meta.get('archive_name', 'smurf')
                try:
                    if (meta.get('format') == 'npy') and (not meta['path'].endswith('.npy')):
                        meta['path'] += '.npy'
                    local_path = meta['path']
                    remote_path = create_remote_path(meta, archive_name)

                    # Only delete files that are in timestamped directories
                    # /data/smurf_data/<timestamp> and are not IV, channel
                    # assignment, or tune files. We may want to add more to
                    # this list of "semi-permanent files" later
                    deletable = True
                    if archive_name == 'smurf':
                        if not local_path.split('/')[3].isdigit():
                            deletable = False
                        for key in ["iv", "channel_assignment", "tune"]:
                            if key in local_path:
                                deletable = False

                    files.append(
                        create_file(local_path, remote_path, archive_name,
                                    deletable=deletable)
                    )
                except Exception as e:
                    self.agent.log.error(
                        "Could not generate SupRsync file object from "
                        "metadata:\n{meta}\nRaised Exception: {e}",
                        meta=meta, e=e
                    )

            if files:
                with srfm.Session.begin() as session:
                    session.add_all(files)

            time.sleep(1)

    def _stop(self, session, params=None):
        self.running = False
        session.set_status('stopping')


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--udp-port', type=int,
                        help="Port for upd-publisher")
    pgroup.add_argument('--create-table', type=bool,
                        help="Specifies whether agent should create or update "
                             "pysmurf_files table if non exists.", default=True)
    pgroup.add_argument('--db-path', type=str, default='/data/so/databases/suprsync.db',
                        help="Path to suprsync sqlite database")
    pgroup.add_argument('--echo-sql', action='store_true')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='PysmurfMonitor',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    monitor = PysmurfMonitor(agent, args)

    agent.register_process('run', monitor.run, monitor._stop, startup=True)

    reactor.listenUDP(args.udp_port, monitor)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import subprocess
import time

import txaio
from ocs import ocs_agent, site_config

from socs.db.suprsync import SupRsyncFileHandler, SupRsyncFilesManager


class SupRsync:
    """
    Agent to rsync files to a remote (or local) destination, verify successful
    transfer, and delete local files after a specified amount of time.

    Parameters
    --------------
    agent : OCSAgent
        OCS agent object
    args : Namespace
        Namespace with parsed arguments

    Attributes
    ------------
    agent : OCSAgent
        OCS agent object
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    archive_name : string
        Name of the managed archive. Sets which files in the suprsync db should
        be copied.
    ssh_host : str, optional
        Remote host to copy data to. If None, will copy data locally.
    ssh_key : str, optional
        ssh-key to use to access the ssh host.
    remote_basedir : path
        Base directory on the destination server to copy files to
    db_path : path
        Path of the sqlite db to monitor
    delete_after : float
        Seconds after which this agent will delete successfully copied files.
    cmd_timeout : float
        Time (sec) for which cmds run on the remote will timeout
    copy_timeout : float
        Time (sec) after which a copy command will timeout
    """

    def __init__(self, agent, args):
        self.agent = agent
        self.log = txaio.make_logger()
        self.archive_name = args.archive_name
        self.ssh_host = args.ssh_host
        self.ssh_key = args.ssh_key
        self.remote_basedir = args.remote_basedir
        self.db_path = args.db_path
        self.delete_after = args.delete_local_after
        self.max_copy_attempts = args.max_copy_attempts
        self.running = False
        self.cmd_timeout = args.cmd_timeout
        self.copy_timeout = args.copy_timeout
        self.files_per_batch = args.files_per_batch
        self.sleep_time = args.sleep_time
        self.compression = args.compression
        self.bwlimit = args.bwlimit

        # Feed for counting transfer errors, loop iterations.
        self.agent.register_feed('transfer_stats',
                                 record=True,
                                 agg_params={
                                     'exclude_aggregator': True,
                                 })

    def run(self, session, params=None):
        """run()

        **Process** - Main run process for the SupRsync agent. Continuosly
        checks the suprsync db checking for files that need to be handled.

        Notes:

            The session data object contains info about recent
            transfers::

                >>> response.session['data']
                {
                  "activity": "idle",
                  "timestamp": 1661284493.5398622,
                  "last_copy": {
                    "start_time": 1661284493.5333128,
                    "files": [],
                    "stop_time": 1661284493.5398622
                  },
                  "counters": {
                    "iterations": 1,
                    "copies": 0,
                    "errors_timeout": 0,
                    "errors_nonzero": 0
                  },
                }
        """

        srfm = SupRsyncFilesManager(self.db_path, create_all=True)

        handler = SupRsyncFileHandler(
            srfm, self.archive_name, self.remote_basedir, ssh_host=self.ssh_host,
            ssh_key=self.ssh_key, cmd_timeout=self.cmd_timeout,
            copy_timeout=self.copy_timeout, compression=self.compression,
            bwlimit=self.bwlimit
        )

        self.running = True
        session.set_status('running')

        # Note this will also be stored directly in session.data
        counters = {
            'iterations': 0,
            'copies': 0,
            'errors_timeout': 0,
            'errors_nonzero': 0,
        }

        session.data = {
            'activity': 'idle',
            'last_copy': {},
            'counters': counters,
        }

        next_feed_update = 0

        while self.running:
            counters['iterations'] += 1

            op = {'start_time': time.time()}
            try:
                session.data['activity'] = 'copying'
                op['files'] = handler.copy_files(max_copy_attempts=self.max_copy_attempts,
                                                 num_files=self.files_per_batch)
                counters['copies'] += len(op['files'])
            except subprocess.TimeoutExpired as e:
                self.log.error("Timeout when copying files! {e}", e=e)
                op['error'] = 'timed out'
                counters['errors_timeout'] += 1
            except subprocess.CalledProcessError as e:
                self.log.error("rsync returned non-zero exit code! {e}", e=e)
                op['error'] = 'nonzero exit'
                counters['errors_nonzero'] += 1

            now = time.time()
            op['stop_time'] = now
            session.data['last_copy'] = op
            session.data['timestamp'] = now

            if now >= next_feed_update:
                self.agent.publish_to_feed('transfer_stats', {
                    'block_name': 'block0',
                    'timestamp': now,
                    'data': counters})
                next_feed_update = now + 10 * 60

            if self.delete_after is not None:
                session.data['activity'] = 'deleting'
                handler.delete_files(self.delete_after)

            session.data['activity'] = 'idle'
            time.sleep(self.sleep_time)

        return True, "Stopped run process"

    def _stop(self, session, params=None):
        self.running = False
        session.set_status('stopping')


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--archive-name', required=True, type=str,
                        help="Name of managed archive. Determines which files "
                             "should be copied")
    pgroup.add_argument('--remote-basedir', required=True, type=str,
                        help="Base directory on the remote server where files "
                             "will be copied")
    pgroup.add_argument('--db-path', required=True, type=str,
                        help="Path to the suprsync sqlite db")
    pgroup.add_argument('--ssh-host', type=str, default=None,
                        help="Remote host to copy files to (e.g. "
                             "'<user>@<host>'). If None, will copy files locally")
    pgroup.add_argument('--ssh-key', type=str,
                        help="Path to ssh-key needed to access remote host")
    pgroup.add_argument('--delete-local-after', type=float,
                        help="Time (sec) after which this agent will delete "
                             "local copies of successfully transfered files. "
                             "If None, will not delete files.")
    pgroup.add_argument('--max-copy-attempts', type=int, default=5,
                        help="Number of failed copy attempts before the agent "
                             "will stop trying to copy a file")
    pgroup.add_argument('--copy-timeout', type=float,
                        help="Time (sec) before the rsync command will timeout")
    pgroup.add_argument('--cmd-timeout', type=float,
                        help="Time (sec) before remote commands will timeout")
    pgroup.add_argument('--files-per-batch', type=int,
                        help="Number of files to copy over per batch. Default "
                        "is None, which will copy over all available files.")
    pgroup.add_argument('--sleep-time', type=float, default=60,
                        help="Time to sleep (sec) in between copy iterations")
    pgroup.add_argument('--compression', action='store_true', default=False,
                        help="Activate gzip on data transfer (rsync -z)")
    pgroup.add_argument('--bwlimit', type=str, default=None,
                        help="Bandwidth limit arg (passed through to rsync)")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args('SupRsync',
                                  parser=parser,
                                  args=args)
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    agent, runner = ocs_agent.init_site_agent(args)
    suprsync = SupRsync(agent, args)
    agent.register_process('run', suprsync.run, suprsync._stop, startup=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import calendar
import select
import socket
import struct
import time
from collections import deque

import numpy as np
import txaio

txaio.use_twisted()

# should be consistent with the software on beaglebone
COUNTER_INFO_LENGTH = 100
# header, quad, clock[100], clock_overflow[100], refcount[100], error[100]
COUNTER_PACKET_SIZE = 4 + 4 * 5 * COUNTER_INFO_LENGTH
# header, clock, clock_overflow, info[10], synch[10], synch_overflow[10]
IRIG_PACKET_SIZE = 132
# header, type
TIMEOUT_PACKET_SIZE = 8


class EncoderParser:

    def __init__(self, beaglebone_port=50007, read_chunk_size=8192):

        self.encoder_queue = deque()
        self.irig_queue = deque()

        self.is_start = 1
        self.start_time = [0, 0, 0]
        self.current_time = 0

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(('', beaglebone_port))

        self.data = ''
        self.read_chunk_size = read_chunk_size

        self.log = txaio.make_logger()

    def check_once(self):

        enc_oncetime = False
        irig_oncetime = False
        checking_start = time.time()
        sampling_timeout = 30

        while True:
            ready = select.select([self.sock], [], [], 2)
            if ready[0]:
                self.data = self.sock.recv(self.read_chunk_size)
            if not self.check_data_length(0, 4):
                self.log.error(
                    'Header data length error in check_once()')
            header = self.data[0:4]
            header = struct.unpack('<I', header)[0]

            if enc_oncetime and irig_oncetime:
                print('---finished taking each data---')
                break
            if header == 0x1eaf and not enc_oncetime:  # Encoder data
                if not self.check_data_length(0, COUNTER_PACKET_SIZE):
                    self.log.error(
                        'Failed to catch the Encoder data')
                self.parse_counter_info(self.data[4: COUNTER_PACKET_SIZE])
                res_bb = self.encoder_queue.popleft()
                print('---Encoder Response---')
                print(res_bb)
                enc_oncetime = True
            elif header == 0xcafe and not irig_oncetime:
                if not self.check_data_length(0, IRIG_PACKET_SIZE):
                    self.log.error(
                        'Failed to catch the IRIG data')
                self.parse_irig_info(self.data[4:IRIG_PACKET_SIZE])
                res_irig = self.irig_queue.popleft()
                print('---IRIG Response---')
                print(res_irig)
                irig_oncetime = True
            else:
                pass

            if time.time() - checking_start > sampling_timeout:
                print('time out')
                break
        # End of check_once()

    def check_data_length(self, start_index, size_of_read):
        if start_index + size_of_read > len(self.data):
            self.data = self.data[start_index:]
            return False
        return True

    def grab_and_parse_data(self):
        while True:
            ready = select.select([self.sock], [], [], 2)
            if ready[0]:
                data = self.sock.recv(self.read_chunk_size)
                if len(self.data) > 0:
                    self.data += data
                else:
                    self.data = data

                while True:
                    if not self.check_data_length(0, 4):
                        self.log.error(
                            'Header data length error in grab_and_parse_data()'
                        )
                        break

                    header = self.data[0:4]
                    header = struct.unpack('<I', header)[0]

                    # 0x1EAF = Encoder Packet
                    # 0xCAFE = IRIG Packet
                    # 0xE12A = Error Packet

                    # Encoder
                    if header == 0x1eaf:
                        if not self.check_data_length(0, COUNTER_PACKET_SIZE):
                            self.log.error(
                                'Failed to catch the Encoder data')
                            break
                        self.parse_counter_info(
                            self.data[4: COUNTER_PACKET_SIZE])
                        if len(self.data) >= COUNTER_PACKET_SIZE:
                            self.data = self.data[COUNTER_PACKET_SIZE:]

                    # IRIG
                    elif header == 0xcafe:
                        if not self.check_data_length(0, IRIG_PACKET_SIZE):
                            self.log.error(
                                'Failed to catch the IRIG data')
                            break
                        self.parse_irig_info(self.data[4: IRIG_PACKET_SIZE])
                        if len(self.data) >= IRIG_PACKET_SIZE:
                            self.data = self.data[IRIG_PACKET_SIZE:]

                    elif header == 0x1234:
                        if not self.check_data_length(0, TIMEOUT_PACKET_SIZE):
                            self.log.error(
                                'Failed to catch the Timeout data')
                            self.data = ''
                            break
                        timeout_type = self.data[4:TIMEOUT_PACKET_SIZE]
                        timeout_type = struct.unpack('<I', timeout_type)[0]
                        if timeout_type == 1:
                            self.log.error('Recieved Encoder timeout packet.')
                        elif timeout_type == 2:
                            self.log.error('Recieved IRIG timeout packet.')
                        else:
                            self.log.error('Recieved timeout packet but '
                                           'timeout-type(={}) is unknown type.'
                                           .format(timeout_type))
                        self.data = ''
                        break
                    else:
                        self.log.error('Bad header')
                        self.data = ''

                    if len(self.data) == 0:
                        break
                break

    def parse_counter_info(self, data):
        derter = np.array(
            struct.unpack('<' + 'LLLLL' * COUNTER_INFO_LENGTH, data))

        self.encoder_queue.append(
            (derter[0:COUNTER_INFO_LENGTH],
             derter[COUNTER_INFO_LENGTH:2 * COUNTER_INFO_LENGTH]
             + (derter[2 * COUNTER_INFO_LENGTH:3 * COUNTER_INFO_LENGTH] << 32),
             derter[3 * COUNTER_INFO_LENGTH:4 * COUNTER_INFO_LENGTH],
             derter[4 * COUNTER_INFO_LENGTH:5 * COUNTER_INFO_LENGTH],
             time.time()))

    def parse_irig_info(self, data):

        unpacked_data = struct.unpack('<' + 'LL' + 'LLL' * 10, data)
        rising_edge_time = unpacked_data[0] + (unpacked_data[1] << 32)
        irig_info = unpacked_data[2:12]
        irig_time = self.pretty_print_irig_info(irig_info, rising_edge_time)
        synch_pulse_clock_times = (np.asarray(unpacked_data[12:22])
                                   + (np.asarray(unpacked_data[22:32]) << 32)
                                   ).tolist()
        self.irig_queue.append((rising_edge_time, irig_time, irig_info,
                                synch_pulse_clock_times, time.time()))

    def pretty_print_irig_info(self, irig_info, edge, print_out=False):

        secs = self.de_irig(irig_info[0], 1)
        mins = self.de_irig(irig_info[1], 0)
        hours = self.de_irig(irig_info[2], 0)
        day = self.de_irig(irig_info[3], 0)\
            + self.de_irig(irig_info[4], 0) * 100
        year = self.de_irig(irig_info[5], 0)

        if self.is_start == 1:
            self.start_time = [hours, mins, secs]
            self.is_start = 0

        if print_out:

            dsecs = secs - self.start_time[2]
            dmins = mins - self.start_time[1]
            dhours = hours - self.start_time[0]

            if dhours < 0:
                dhours = dhours + 24

            if (dmins < 0) or ((dmins == 0) and (dsecs < 0)):
                dmins = dmins + 60
                dhours = dhours - 1

            if dsecs < 0:
                dsecs = dsecs + 60
                dmins = dmins - 1

            print('Current Time:', ('%d:%d:%d' % (hours, mins, secs)),
                  'Run Time', ('%d:%d:%d' % (dhours, dmins, dsecs)),
                  'Clock Count\n', edge)

        try:
            st_time = time.strptime(
                "%d %d %d:%d:%d" % (year, day, hours, mins, secs),
                "%Y %j %H:%M:%S")
            self.current_time = calendar.timegm(st_time)
        except ValueError:
            self.log.error(f'Invalid IRIG-B timestamp: '
                           f'{year} {day} {hours} {mins} {secs}')
            self.current_time = -1

        return self.current_time

    def de_irig(self, val, base_shift=0):
        return (
            ((val >> (0 + base_shift)) & 1)
            + ((val >> (1 + base_shift)) & 1) * 2
            + ((val >> (2 + base_shift)) & 1) * 4
            + ((val >> (3 + base_shift)) & 1) * 8
            + ((val >> (5 + base_shift)) & 1) * 10
            + ((val >> (6 + base_shift)) & 1) * 20
            + ((val >> (7 + base_shift)) & 1) * 40
            + ((val >> (8 + base_shift)) & 1) * 80
        )

    def __del__(self):
        self.sock.close()


if __name__ == '__main__':
    test = EncoderParser()
    test.check_once()
    del test
